# -*- encoding: utf-8 -*-
# kb v0.1.7
# A knowledge base organizer
# Copyright © 2020, gnc.
# See /LICENSE for licensing information.

"""
kb config module

:Copyright: © 2020, gnc.
:License: GPLv3 (see /LICENSE).
"""

__all__ = ()

import os
from sys import platform
from pathlib import Path
import toml

BASE_PATH = Path(os.environ.get("XDG_DATA_HOME",Path(Path.home(),".local","share")),"kb")


DEFAULT_CONFIG = {
    "PATH_KB": str(Path(BASE_PATH)),
    "PATH_KB_DB": str(Path(BASE_PATH, "kb.db")),
    "PATH_KB_HIST": str(Path(BASE_PATH, "recent.hist")),
    "PATH_KB_DATA": str(Path(BASE_PATH, "data")),
    "PATH_KB_GIT": str(Path(BASE_PATH, ".git")),
    # for future use
    "PATH_KB_CONFIG": str(Path(BASE_PATH, "kb.conf.py")),
    "PATH_KB_TEMPLATES": str(Path(BASE_PATH, "templates")),
    "PATH_KB_DEFAULT_TEMPLATE": str(Path(BASE_PATH, "templates", "default")),
    "DB_SCHEMA_VERSION": 1,
    "EDITOR": os.environ.get("EDITOR", "vim"),
    "INITIAL_CATEGORIES": ["default", ]
}


DEFAULT_TEMPLATE = {
    "TITLES": ("^#.*", "blue"),
    "WARNINGS": ("^!.*", "yellow"),
}


def get_markers(markers_path: str):
    """
    Load markers file

    Arguments:
    markers_path    - the path to the toml markers file

    Returns a dictionary containing markers
    """
    try:
        return toml.load(markers_path)
    except toml.TomlDecodeError:
        print("Error: The provided file is not in the toml format")
    except FileNotFoundError:
        print("Error: The provided file does not exist or cannot be accessed")



# -*- encoding: utf-8 -*-
# kb v0.1.7
# A knowledge base organizer
# Copyright © 2020, gnc.
# See /LICENSE for licensing information.

"""
kb artifact frozen dataclass

:Copyright: © 2020, gnc.
:License: GPLv3 (see /LICENSE).
"""

import attr
from typing import List, Set, Optional


@attr.s(auto_attribs=True, frozen=True, slots=True)
class Artifact:
    id: Optional[int]
    title: str
    category: str
    path: Optional[str] = None
    tags: Optional[str] = None
    status: Optional[str] = None
    author: Optional[str] = None
    template: Optional[str] = None



# -*- encoding: utf-8 -*-
# kb v0.1.7
# A knowledge base organizer
# Copyright © 2020, gnc.
# See /LICENSE for licensing information.

"""
kb export command module

:Copyright: © 2020, gnc.
:License: GPLv3 (see /LICENSE).
"""

import time
import tarfile
from pathlib import Path
from typing import Dict


def export(args: Dict[str, str], config: Dict[str, str]):
    """
    Export the entire kb knowledge base.

    Arguments:
    args:           - a dictionary containing the following fields:
                      file -> a string representing the wished output
                        filename
    config:         - a configuration dictionary containing at least
                      the following keys:
                      PATH_KB           - the main path of KB
    """
    fname = args["file"] or time.strftime("%d_%m_%Y-%H%M%S")
    archive_ext = ".kb.tar.gz"
    if not fname.endswith(archive_ext):
        fname = fname + archive_ext

    if args["only_data"]:
        with tarfile.open(fname, mode='w:gz') as archive:
            archive.add(config["PATH_KB_DATA"], arcname="kb", recursive=True)
    else:
        with tarfile.open(fname, mode='w:gz') as archive:
            archive.add(config["PATH_KB"], arcname=".kb", recursive=True)



# -*- encoding: utf-8 -*-
# kb v0.1.7
# A knowledge base organizer
# Copyright © 2020, gnc.
# See /LICENSE for licensing information.

"""
kb edit command module

:Copyright: © 2020, gnc.
:License: GPLv3 (see /LICENSE).
"""

import shlex
from pathlib import Path
from subprocess import call
from typing import Dict
import kb.db as db
import kb.initializer as initializer
import kb.history as history


def edit(args: Dict[str, str], config: Dict[str, str]):
    """
    Edit the content of an artifact.

    Arguments:
    args:           - a dictionary containing the following fields:
                      id -> the IDs (the one you see with kb list)
                        associated to the artifact we want to edit
                      title -> the title assigned to the artifact(s)
                      category -> the category assigned to the artifact(s)
    config:         - a configuration dictionary containing at least
                      the following keys:
                      PATH_KB_DB        - the database path of KB
                      PATH_KB_DATA      - the data directory of KB
                      PATH_KB_HIST      - the history menu path of KB
                      EDITOR            - the editor program to call
    """
    initializer.init(config)

    # if an ID is specified, load artifact with that ID
    if args["id"]:
        edit_by_id(args["id"], config)

    # else if a title is specified
    elif args["title"]:
        edit_by_name(args["title"], args["category"], config)

    # else try to guess
    elif args["nameid"]:
        if args["nameid"].isdigit():
            edit_by_id(args["nameid"], config)
        else:
            edit_by_name(args["nameid"], args["category"], config)


def edit_by_id(id: int, config: Dict[str, str]):
    """
    Edit the content of an artifact by id.

    Arguments:
    id:             - the ID (the one you see with kb list)
                      associated to the artifact to edit
    config:         - a configuration dictionary containing at least
                      the following keys:
                      PATH_KB_DB        - the database path of KB
                      PATH_KB_DATA      - the data directory of KB
                      PATH_KB_HIST      - the history menu path of KB
                      EDITOR            - the editor program to call
    """
    conn = db.create_connection(config["PATH_KB_DB"])
    artifact = history.get_artifact(
        conn, config["PATH_KB_HIST"], id)

    category_path = Path(config["PATH_KB_DATA"], artifact.category)

    shell_cmd = shlex.split(config["EDITOR"]) + \
        [str(Path(category_path, artifact.title))]
    call(shell_cmd)


def edit_by_name(title: str, category: str, config: Dict[str, str]):
    """
    Edit the content of an artifact by name, that is title/category

    Arguments:
    title:          - the title assigned to the artifact(s)
    category:       - the category assigned to the artifact(s)
    config:         - a configuration dictionary containing at least
                      the following keys:
                      PATH_KB_DB        - the database path of KB
                      PATH_KB_DATA      - the data directory of KB
                      PATH_KB_HIST      - the history menu path of KB
                      EDITOR            - the editor program to call
    """
    conn = db.create_connection(config["PATH_KB_DB"])
    artifacts = db.get_artifacts_by_filter(conn, title=title,
                                           category=category,
                                           is_strict=True)

    if len(artifacts) == 1:
        artifact = artifacts.pop()
        category_path = Path(config["PATH_KB_DATA"], artifact.category)
        shell_cmd = shlex.split(
            config["EDITOR"]) + [str(Path(category_path, artifact.title))]
        call(shell_cmd)
    elif len(artifacts) > 1:
        print(
            "There is more than one artifact with that title, please specify a category")
    else:
        print(
            "There is no artifact with that name, please specify a correct artifact name")



# -*- encoding: utf-8 -*-
# kb v0.1.7
# A knowledge base organizer
# Copyright © 2020, gnc.
# See /LICENSE for licensing information.

"""
Main routine of kb.

:Copyright: © 2020, gnc.
:License: GPLv3 (see /LICENSE).
"""
from kb.main import main


__all__ = ('main',)


if __name__ == '__main__':
    main()



# -*- encoding: utf-8 -*-
# kb v0.1.7
# A knowledge base organizer
# Copyright © 2020, gnc.
# See /LICENSE for licensing information.

"""
kb grep command module

:Copyright: © 2020, gnc.
:License: GPLv3 (see /LICENSE).
"""

import sys
from pathlib import Path
from typing import Dict
import kb.db as db
import kb.initializer as initializer
import kb.printer.grep as printer
import kb.history as history
import kb.filesystem as fs


def grep(args: Dict[str, str], config: Dict[str, str]):
    """
    Grep through the list of artifacts of the knowledge base of kb.

    Arguments:
    args:           - a dictionary containing the following fields:
                      regex -> the regex to search for
                      case_insensitive -> a boolean, if true,
                        the search will be case insensitive
                      matches -> a boolean, if true, only the raw
                        matches will be shown
                      verbose -> a boolean, if true, a verbose
                        output is produced on screen
    config:         - a configuration dictionary containing at least
                      the following keys:
                      PATH_KB_DB        - the database path of KB
                      PATH_KB_DATA      - the data directory of KB
                      PATH_KB_HIST      - the history menu path of KB
    """
    initializer.init(config)

    conn = db.create_connection(config["PATH_KB_DB"])

    # Get all artifacts
    rows = db.get_artifacts_by_filter(conn, title="")

    # Get all the file paths related to the artifacts in the database
    file_list = [Path(config["PATH_KB_DATA"], r.category, r.title)
                 for r in rows]

    # Grep in the files
    results = fs.grep_in_files(
        file_list,
        args["regex"],
        args["case_insensitive"])

    # If user specified --matches -> just show matching lines and exit
    color_mode = not args["no_color"]
    if args["matches"]:
        printer.print_grep_matches(results, color_mode)
        sys.exit(0)

    # Get the list of artifact tuples in the form (category,title)
    artifact_names = [fs.get_filename_parts_wo_prefix(
        res[0], config["PATH_KB_DATA"]) for res in results]

    # Get the set of uniq artifacts
    uniq_artifact_names = set(artifact_names)

    # Get the number of matches (hits) for each path found
    filecounts = get_hits_per_artifact_name(artifact_names)

    grep_result = list()

    for art in uniq_artifact_names:

        artifact = db.get_artifacts_by_filter(
            conn, category="/".join(art[:-1]), title=art[-1], is_strict=True)[0]

        if artifact:
            no_of_hits = filecounts[art]
            grep_result.append((artifact, no_of_hits))

    # Sort by number of hits, the largest -> the first
    grep_result.sort(key=lambda x: x[1], reverse=True)

    grep_artifacts = [r[0] for r in grep_result]
    grep_hits = [r[1] for r in grep_result]

    # Write to history file
    history.write(config["PATH_KB_HIST"], grep_artifacts)

    color_mode = not args["no_color"]
    if args["verbose"]:
        printer.print_grep_result_verbose(
            grep_artifacts, grep_hits, color_mode)
    else:
        printer.print_grep_result(grep_artifacts, grep_hits, color_mode)


def get_hits_per_artifact_name(artifact_name_list):
    """
    Get the dictionary related to the number of hits
    for each artifact name found.

    Arguments:
    artifact_name_list  - a list of tuples, where each
                        tuple is (category, title)
    Returns:
    A dictionary having as key the tuple (category,title)
    and as value the number of hits of that
    """
    namecounts = dict()
    for artname in artifact_name_list:
        if artname in namecounts:
            namecounts[artname] += 1
        else:
            namecounts[artname] = 1
    return namecounts



# -*- encoding: utf-8 -*-
# kb v0.1.7
# A knowledge base organizer
# Copyright © 2020, gnc.
# See /LICENSE for licensing information.

"""
kb template command module

:Copyright: © 2020, gnc.
:License: GPLv3 (see /LICENSE).
"""

import shlex
import sys
import toml
from pathlib import Path
from subprocess import call
from typing import Dict, List
import kb.db as db
import kb.initializer as initializer
import kb.filesystem as fs
import kb.config as conf
from kb.entities.artifact import Artifact
import kb.printer.template as printer


def get_templates(templates_path: str) -> List[str]:
    """
    Get the list of available templates.

    Arguments:
    templates_path      - the path where all templates are stored

    Returns:
    A list of strings representing the available templates
    """
    return fs.list_files(templates_path)


def search(args: Dict[str, str], config: Dict[str, str]):
    """
    Search templates installed in kb.

    Arguments:
    args:           - a dictionary containing the following fields:
                      query -> filter for the title field of the artifact
    config:         - a configuration dictionary containing at least
                      the following keys:
                      PATH_KB_TEMPLATES     - the path to where the templates of KB
                                              are stored
    """
    # template_list = fs.list_files(config["PATH_KB_TEMPLATES"])
    template_list = get_templates(config["PATH_KB_TEMPLATES"])
    if not template_list:
        return

    if args["query"]:
        template_list = [x for x in template_list if args["query"] in x]
    color_mode = not args["no_color"]
    printer.print_template_search_result(template_list, color_mode)


def apply_on_set(args: Dict[str, str], config: Dict[str, str]):
    """
    Apply the specified template to all the filtered artifacts
    """
    # Check initialization
    initializer.init(config)

    tags_list = None
    if args["tags"] and args["tags"] != "":
        tags_list = args["tags"].split(';')

    conn = db.create_connection(config["PATH_KB_DB"])
    is_query_strict = not args["extended_match"]
    rows = db.get_artifacts_by_filter(
        conn,
        title=args["title"],
        category=args["category"],
        tags=tags_list,
        status=args["status"],
        author=args["author"],
        is_strict=is_query_strict)

    for artifact in rows:
        updated_artifact = Artifact(
            id=artifact.id,
            title=artifact.title,
            category=artifact.category,
            tags=artifact.tags,
            author=artifact.author,
            status=artifact.status,
            template=args["template"])
        db.update_artifact_by_id(conn, artifact.id, updated_artifact)


def new(args: Dict[str, str], config: Dict[str, str]):
    """
    Create a new template from scratch starting from the default template.

    Arguments:
    args:           - a dictionary containing the following fields:
                      template -> the name of the new template to create
    config:         - a configuration dictionary containing at least
                      the following keys:
                      PATH_KB_TEMPLATES         - the path to where the templates of KB
                                                  are stored
                      PATH_KB_DEFAULT_TEMPLATE  - the path to where the default template of KB
                                                  is stored
                      EDITOR                    - the editor program to call
    """
    template_path = str(Path(config["PATH_KB_TEMPLATES"]) / args["template"])

    if fs.is_file(template_path):
        print(
            "ERROR: The template you inserted corresponds to an existing one. "
            "Please specify another name for the new template")
        sys.exit(1)

    fs.create_directory(Path(template_path).parent)
    # fs.copy_file(config["PATH_KB_DEFAULT_TEMPLATE"], template_path)

    with open(template_path, 'w') as tmplt:
        tmplt.write("# This is an example configuration template\n\n\n")
        tmplt.write(toml.dumps(conf.DEFAULT_TEMPLATE))

    shell_cmd = shlex.split(
        config["EDITOR"]) + [template_path]
    call(shell_cmd)


def add(args: Dict[str, str], config: Dict[str, str]):
    """
    Add a new template to the templates available in kb.

    Arguments:
    args:           - a dictionary containing the following fields:
                      file -> the path to the template to include in kb templates
                      title -> the title to assign to the kb template added
    config:         - a configuration dictionary containing at least
                      the following keys:
                      PATH_KB_TEMPLATES         - the path to where the templates of KB
                                                  are stored
    """
    template_path = args["file"]
    if args["title"]:
        dest_path = str(Path(config["PATH_KB_TEMPLATES"]) / args["title"])
    else:
        dest_path = config["PATH_KB_TEMPLATES"]
    fs.copy_file(template_path, dest_path)


def delete(args: Dict[str, str], config: Dict[str, str]):
    """
    Delete a template from the kb templates.

    Arguments:
    args:           - a dictionary containing the following fields:
                      template -> the name of the template to remove
    config:         - a configuration dictionary containing at least
                      the following keys:
                      PATH_KB_TEMPLATES         - the path to where the templates of KB
                                                  are stored
    """
    template_name = args["template"]
    fs.remove_file(Path(config["PATH_KB_TEMPLATES"], template_name))


def edit(args: Dict[str, str], config: Dict[str, str]):
    """
    Edit a template from the kb templates.

    Arguments:
    args:           - a dictionary containing the following fields:
                      template -> the name of the template to edit
    config:         - a configuration dictionary containing at least
                      the following keys:
                      PATH_KB_TEMPLATES  - the path to where the templates of KB
                                           are stored
                      EDITOR             - the editor program to call
    """
    template_path = str(Path(config["PATH_KB_TEMPLATES"]) / args["template"])

    if not fs.is_file(template_path):
        print("ERROR: The template you want to edit does not exist. "
              "Please specify a valid template to edit or create a new one")
        sys.exit(1)

    shell_cmd = shlex.split(
        config["EDITOR"]) + [template_path]
    call(shell_cmd)


COMMANDS = {
    'add': add,
    'delete': delete,
    'edit': edit,
    'list': search,
    'new': new,
    'apply': apply_on_set,
}


def template(args: Dict[str, str], config: Dict[str, str]):
    """
    Manage templates for kb.

    Arguments:
    args:           - a dictionary containing the following fields:
                      template_command -> the sub-command to execute for templates
                                          that can be: "add", "delete", "edit",
                                          "list" or "new".
                      file -> used if the command is add, representing the template
                              file to add to kb
                      template -> used if the command is "delete", "edit" or "new"
                                  to represent the name of the template
                      query -> used if the command is "list"
    config:         - a configuration dictionary containing at least
                      the following keys:
                      PATH_KB_DEFAULT_TEMPLATE - the path to the kb default template
                      PATH_KB_TEMPLATES        - the path to kb templates
                      EDITOR                   - the editor program to call
    """

    # Check initialization
    initializer.init(config)

    COMMANDS[args["template_command"]](args, config)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# kb test suite
# Copyright © 2020, gnc.
# See /LICENSE for licensing information.

import os
import kb
import kb.filesystem as fs
from pathlib import Path


def test_list_files():
    # Test number of files found
    files_kb = fs.list_files(Path("tests", "data", ".kb", "data"))
    assert len(files_kb) == 4

    files_kb2 = fs.list_files(Path("tests", "data", ".kb2", "data"))
    assert len(files_kb2) == 8
    assert not len(files_kb2) == 7


    # Test if files picked are right
    kb_list_of_files = ["pt_tls",
                        "pt_ipmi",
                        "pth",
                        "pt_wifi_tips"]

    assert set(files_kb) == set(kb_list_of_files)
    
    kb2_list_of_files = [str(Path("dir2","dir3","pt_wifi2")),
                         str(Path("dir2","pt_wifi_tips")),
                         str(Path("pt_tls")),
                         str(Path("pt_ipmi")),
                         str(Path("pth")),
                         str(Path("dir1","pt_ipmi")),
                         str(Path("dir1","pth")),
                         str(Path("pt_wifi_tips"))]

    assert set(files_kb2) == set(kb2_list_of_files)

def test_list_dirs():
    # Test number of directories found
    dir_kb = fs.list_dirs(Path("tests", "data", ".kb", "data"))
    assert len(dir_kb) == 0

    dir_kb2 = fs.list_dirs(Path("tests", "data", ".kb2", "data"))
    assert len(dir_kb2) == 3


    # Test if files picked are right
    kb_list_of_dirs =  []

    assert set(dir_kb) == set(kb_list_of_dirs)
    
    kb2_list_of_dirs = [str(Path("dir2")),
                        str(Path("dir2/dir3")),
                        str(Path("dir1"))]

    assert set(dir_kb2) == set(kb2_list_of_dirs)



def test_copy_file():
    filename_src = Path("tests","data","sample_data")
    filename_dst = Path("tests","data","sample_data_dest")

    with open(filename_src, 'w') as f:
        f.write('sample data\n')
    assert os.path.exists(filename_src)

    fs.copy_file(filename_src, filename_dst)
    assert os.path.exists(filename_dst)

def test_move_file():
    filename = Path("tests","data","sample_data")
    with open(filename, 'w') as f:
        f.write('sample data\n')

    assert os.path.exists(filename)

    filename_new = Path("tests","data","new_sample_data")
    fs.move_file(filename, filename_new )
    assert not os.path.exists(filename)
    assert os.path.exists(filename_new)


def test_remove_file():
    filename = Path("tests","data","sample_data")
    with open(filename, 'w') as f:
        f.write('sample data\n')

    assert os.path.exists(filename)

    fs.remove_file(filename)
    assert not os.path.exists(filename)

def test_get_basename():
    filename = Path("tests","data","sample_data")
    with open(filename, 'w') as f:
        f.write('sample data\n')
    assert "sample_data" == fs.get_basename(filename)




# -*- encoding: utf-8 -*-
# kb v0.1.7
# A knowledge base organizer
# Copyright © 2020, gnc.
# See /LICENSE for licensing information.

"""
Command Line Parsing Module for kb

:Copyright: © 2020, gnc.
:License: GPLv3 (see /LICENSE).
"""

__all__ = ()

import sys
import argparse
from kb import __version__
from typing import Sequence


def parse_args(args: Sequence[str]) -> argparse.Namespace:
    """
    This function parses the arguments which have been passed from the command
    line, these can be easily retrieved for example by using "sys.argv[1:]".
    It returns an argparse Namespace object.

    Arguments:
    args -- the list of arguments passed from the command line as the sys.argv
            format

    Returns:
    An argparse Namespace object with the provided arguments, which
    can be used in a simpler format.
    """
    parser = argparse.ArgumentParser(prog='kb',
                                     description='A knowledge base organizer')

    parser.add_argument(
        "--version",
        action="version",
        version="%(prog)s {}".format(__version__))

    subparsers = parser.add_subparsers(help='commands', dest="command")
    subparsers.required = True

    # Main Commands
    add_parser = subparsers.add_parser(
        'add', help='Add an artifact')
    edit_parser = subparsers.add_parser(
        'edit', help='Edit an artifact content')
    list_parser = subparsers.add_parser(
        'list', help='Search for artifacts')
    view_parser = subparsers.add_parser(
        'view', help='View artifacts')
    grep_parser = subparsers.add_parser(
        'grep', help='Grep through kb artifacts')
    update_parser = subparsers.add_parser(
        'update', help='Update artifact properties')
    delete_parser = subparsers.add_parser(
        'delete', help='Delete artifacts')
    template_parser = subparsers.add_parser(
        'template', help='Manage templates for artifacts')
    import_parser = subparsers.add_parser(
        'import', help='Import a knowledge base')
    export_parser = subparsers.add_parser(
        'export', help='Export the knowledge base')
    erase_parser = subparsers.add_parser(
        'erase', help='Erase the entire kb knowledge base')
    sync_parser = subparsers.add_parser(
        'sync', help='Synchronize the knowledge base with a remote git repository')
    help_parser = subparsers.add_parser(
        'help', help='Show help of a particular command')

    # add parser
    add_parser.add_argument(
        "file",
        help="Path of the file to add to kb as artifact",
        type=str,
        nargs="*",
    )
    add_parser.add_argument(
        "-t", "--title",
        help="Title of the added artifact",
        type=str,
    )
    add_parser.add_argument(
        "-c", "--category",
        help="Category associated to the artifact",
        default="default",
        type=str,
    )
    add_parser.add_argument(
        "-g", "--tags",
        help="""
        Tags to associate to the artifact in the form \"tag1;tag2;...;tagN\"
        """,
        type=str,
    )
    add_parser.add_argument(
        "-a", "--author",
        help="Author of the artifact",
        type=str,
    )
    add_parser.add_argument(
        "-s", "--status",
        help="Status of the artifact",
        type=str,
    )
    add_parser.add_argument(
        "--template",
        help="Template to apply to the artifact",
        type=str,
    )
    add_parser.add_argument(
        "-b", "--body",
        help="Body of the artifact",
        type=str,
    )

    # edit parser
    edit_parser.add_argument(
        "nameid",
        help="Title or ID of the artifact to edit",
        type=str,
        nargs="?",
    )
    edit_parser.add_argument(
        "-i", "--id",
        help="ID of the artifact to edit",
        type=str,
    )
    edit_parser.add_argument(
        "-t", "--title",
        help="Title to update",
        default=None,
        type=str,
    )
    edit_parser.add_argument(
        "-c", "--category",
        help="Category to update",
        default=None,
        type=str,
    )
    edit_parser.add_argument(
        "-g", "--tags",
        help="Tags to update in the form \"tag1;tag2;...;tagN\"",
        default=None,
        type=str,
    )
    edit_parser.add_argument(
        "-a", "--author",
        help="Author to update",
        default=None,
        type=str,
    )
    edit_parser.add_argument(
        "-s", "--status",
        help="Status to update",
        default=None,
        type=str,
    )

    # list parser
    list_parser.add_argument(
        "query",
        help="Filter search results by specified title",
        default="",
        nargs="?",
        type=str,
    )
    list_parser.add_argument(
        "-c", "--category",
        help="Filter search results by specified category",
        default=None,
        type=str,
    )
    list_parser.add_argument(
        "-g", "--tags",
        help="""
        Tags associates to the artifact to search in the form \"tag1;tag2;...;tagN\"
        """,
        default=None,
        type=str,
    )
    list_parser.add_argument(
        "-a", "--author",
        help="Filter search results by specified author",
        default=None,
        type=str,
    )
    list_parser.add_argument(
        "-s", "--status",
        help="Filter search results by specified status",
        default=None,
        type=str,
    )
    list_parser.add_argument(
        "-v", "--verbose",
        help="Show additional information for the provided results",
        action='store_true',
        dest='verbose',
        default=False,
    )
    list_parser.add_argument(
        "-f", "--full-identifier",
        help="Print results in full-identifier mode",
        action='store_true',
        dest='full_identifier',
        default=False,
    )
    list_parser.add_argument(
        "-n", "--no-color",
        help="Enabled no-color mode",
        action='store_true',
        dest='no_color',
        default=False,
    )

    # view parser
    view_parser.add_argument(
        "nameid",
        help="Title or ID of the artifact to view",
        type=str,
        nargs="?",
    )
    view_parser.add_argument(
        "-i", "--id",
        help="ID of the artifact to visualize",
        type=str,
    )
    view_parser.add_argument(
        "-t", "--title",
        help="Title of the artifact to visualize",
        type=str,
    )
    view_parser.add_argument(
        "-c", "--category",
        help="Category associated to the artifact to visualize",
        type=str,
    )
    view_parser.add_argument(
        "-e", "--open-editor",
        help="Open the file in a text editor (read-only mode)",
        action='store_true',
        dest='editor',
        default=False,
    )
    view_parser.add_argument(
        "-n", "--no-color",
        help="Enabled no-color mode",
        action='store_true',
        dest='no_color',
        default=False,
    )

    # grep parser
    grep_parser.add_argument(
        "regex",
        help="Filter search results by specified regex",
        type=str,
    )
    grep_parser.add_argument(
        "-c", "--category",
        help="Filter search results by specified category",
        default=None,
        type=str,
    )
    grep_parser.add_argument(
        "-g", "--tags",
        help="""
        Tags associates to the artifact to search in the form \"tag1;tag2;...;tagN\"
        """,
        default=None,
        type=str,
    )
    grep_parser.add_argument(
        "-a", "--author",
        help="Filter search results by specified author",
        default=None,
        type=str,
    )
    grep_parser.add_argument(
        "-s", "--status",
        help="Filter search results by specified status",
        default=None,
        type=str,
    )
    grep_parser.add_argument(
        "-m", "--show-matches",
        help="Show text matching the regex within the artifact ",
        action='store_true',
        dest='matches',
        default=False,
    )
    grep_parser.add_argument(
        "-i", "--case-insensitive",
        help="Perform grep using a case insensitive regex",
        action='store_true',
        dest='case_insensitive',
        default=False,
    )
    grep_parser.add_argument(
        "-v", "--verbose",
        help="Show additional information for the provided results",
        action='store_true',
        dest='verbose',
        default=False,
    )
    grep_parser.add_argument(
        "-n", "--no-color",
        help="Enabled no-color mode",
        action='store_true',
        dest='no_color',
        default=False,
    )

    # update parser
    update_parser.add_argument(
        "-i", "--id",
        help="ID of the artifact to update",
        type=str,
    )
    update_parser.add_argument(
        "-t", "--title",
        help="Title to update",
        default=None,
        type=str,
    )
    update_parser.add_argument(
        "-c", "--category",
        help="Category to update",
        default=None,
        type=str,
    )
    update_parser.add_argument(
        "-g", "--tags",
        help="Tags to update in the form \"tag1;tag2;...;tagN\"",
        default=None,
        type=str,
    )
    update_parser.add_argument(
        "-a", "--author",
        help="Author to update",
        default=None,
        type=str,
    )
    update_parser.add_argument(
        "-s", "--status",
        help="Status to update",
        default=None,
        type=str,
    )
    update_parser.add_argument(
        "--template",
        help="Template to update",
        default=None,
        type=str,
    )
    update_parser.add_argument(
        "-e", "--edit-content",
        help="Edit content of the artifact with an editor",
        action="store_true",
        dest="edit_content",
    )
    update_parser.add_argument(
        "-b", "--body",
        help="Update the body of the artifact (erases the current content)",
        type=str,
    )

    # delete parser
    delete_parser.add_argument(
        "-i", "--id",
        help="ID of the artifact",
        type=str,
        nargs='*',
    )
    delete_parser.add_argument(
        "-t", "--title",
        help="Title of the artifact to remove",
        default=None,
        type=str,
    )
    delete_parser.add_argument(
        "-c", "--category",
        help="Category associated to the artifact to remove",
        default=None,
        type=str,
    )
    delete_parser.add_argument(
        "-f", "--force",
        help="Force removal without asking for confirmation prompt",
        action='store_true',
        default=False,
    )

    # template parser
    template_subparsers = template_parser.add_subparsers(
        help='template commands', dest="template_command")
    template_subparsers.required = True

    # template subcommands
    add_template_parser = template_subparsers.add_parser(
        'add', help='Add a template from a file')
    edit_template_parser = template_subparsers.add_parser(
        'edit', help='Edit a template')
    list_template_parser = template_subparsers.add_parser(
        'list', help='List all templates')
    new_template_parser = template_subparsers.add_parser(
        'new', help='Create a template from starting from an example')
    delete_template_parser = template_subparsers.add_parser(
        'delete', help='Delete an existing template')
    apply_template_parser = template_subparsers.add_parser(
        'apply', help='Apply a template to an entire set of artifacts')

    add_template_parser.add_argument(
        "file",
        help="The template file to add to kb",
        type=str,
    )
    add_template_parser.add_argument(
        "-t", "--title",
        help="The title to assign to the template added from a file to kb",
        type=str,
    )
    edit_template_parser.add_argument(
        "template",
        help="The name of the template to edit",
        type=str,
    )
    list_template_parser.add_argument(
        "query",
        help="The name (or part of it) of the template to search for",
        type=str,
        nargs='?',
    )
    list_template_parser.add_argument(
        "-n", "--no-color",
        help="Enabled no-color mode",
        action='store_true',
        dest='no_color',
        default=False,
    )
    delete_template_parser.add_argument(
        "template",
        help="The name of the template to delete",
        type=str,
    )
    new_template_parser.add_argument(
        "template",
        help="The name of the template to create",
        type=str,
    )

    apply_template_parser.add_argument(
        "template",
        help="The name of the template to apply to the filtered artifacts",
        type=str,
    )

    apply_template_parser.add_argument(
        "-t", "--title",
        help="Title of the artifacts on which template is applied",
        type=str,
    )
    apply_template_parser.add_argument(
        "-c", "--category",
        help="Category of the artifacts on which template is applied",
        default=None,
        type=str,
    )
    apply_template_parser.add_argument(
        "-g", "--tags",
        help="""
        Tags associates to the artifacts in the form \"tag1;tag2;...;tagN\" where template is applied
        """,
        default=None,
        type=str,
    )
    apply_template_parser.add_argument(
        "-a", "--author",
        help="Author of the artifacts on which template is applied",
        default=None,
        type=str,
    )
    apply_template_parser.add_argument(
        "-s", "--status",
        help="Status of the artifacts on which template is applied",
        default=None,
        type=str,
    )
    apply_template_parser.add_argument(
        "-m", "--extended-match",
        help="""
        Perform application query not on a strict match,
        for example:
        `kb template apply --category cheat -m`
        will match all artifacts containing in their category \"cheat\",
        hence \"cheatsheet\", \"mycheats\",\"cheatsheets\" and so on"
        """,
        action='store_true',
        dest='extended_match',
        default=False,
    )

    # import parser
    import_parser.add_argument(
        "file",
        help="Archive to import as knowledge base",
        type=str,
    )

    # export parser
    export_parser.add_argument(
        "-f", "--file",
        help="Name of the exported archive",
        type=str,
        nargs="?"
    )
    export_parser.add_argument(
        "-d",
        "--only-data",
        help="Export only notes files organized as directories (one for each category)",
        action='store_true',
        dest='only_data',
        default=False,
    )

    # erase parser
    erase_parser.add_argument(
        "--db",
        help="Only remove kb database",
        action='store_true',
        dest='db',
        default=False,
    )

    # sync parser
    sync_parser.add_argument(
        'operation',
        help="""Use \"init\" to initialize the remote repo,
                Use \"push\" to git push (write local -> remote) the knowledge base,
                Use \"pull\" to git pull (retrieve remote -> local) the remote kb,
                Use \"info\" to show information about the repository
             """,
        choices=['init', 'push', 'pull', 'info'])

    help_parser.add_argument(
        'cmd',
        help='Name of command to get help for',
        nargs='?'
    )

    if len(args) == 0:
        parser.print_help(sys.stderr)
        sys.exit(1)

    parsed_args = parser.parse_args()
    if parsed_args.command == 'help':
        if not parsed_args.cmd:
            parser.print_help(sys.stderr)
        else:
            try:
                subparsers.choices[parsed_args.cmd].print_help()
            except KeyError:
                print(f'Unknown command name `{parsed_args.cmd}`')
                print(
                    f"Valid commands are: {', '.join(subparsers.choices.keys())}"
                )
        sys.exit(1)

    return parsed_args



# -*- encoding: utf-8 -*-
# kb v0.1.7
# A knowledge base organizer
# Copyright © 2020, gnc.
# See /LICENSE for licensing information.

"""
kb import command module

:Copyright: © 2020, gnc.
:License: GPLv3 (see /LICENSE).
"""

import tarfile
from pathlib import Path
from typing import Dict
import kb.filesystem as fs


def ingest(args: Dict[str, str], config: Dict[str, str]):
    """
    Import an entire kb knowledge base.

    Arguments:
    args:           - a dictionary containing the following fields:
                      file -> a string representing the path to the archive
                        to be imported
    config:         - a configuration dictionary containing at least
                      the following keys:
                      PATH_KB           - the main path of KB
    """
    if args["file"].endswith(".tar.gz"):
        answer = input("You are about to import a whole knowledge base "
                       "are you sure you want to wipe your previous "
                       " kb data ? [YES/NO]")
        if answer.lower() == "yes":
            print("Previous kb knowledge base data wiped...")
            try:
                fs.remove_directory(config["PATH_KB"])
            except FileNotFoundError:
                pass
            tar = tarfile.open(args["file"], "r:gz")
            tar.extractall(Path.home())
            tar.close()
            print("kb archive {fname} imported".format(fname=args["file"]))
    else:
        print("Please provide a file exported through kb with kb.tar.gz extension")



# -*- encoding: utf-8 -*-
# kb v0.1.7
# A knowledge base organizer
# Copyright © 2020, gnc.
# See /LICENSE for licensing information.

"""
kb delete command module

:Copyright: © 2020, gnc.
:License: GPLv3 (see /LICENSE).
"""

import sys
from typing import Dict
from pathlib import Path
import kb.db as db
import kb.initializer as initializer
import kb.history as history
import kb.filesystem as fs


def delete(args: Dict[str, str], config: Dict[str, str]):
    """
    Delete a list of artifacts from the kb knowledge base.

    Arguments:
    args:           - a dictionary containing the following fields:
                      id -> a list of IDs (the ones you see with kb list)
                        associated to the artifacts we want to delete
                      title -> the title assigned to the artifact(s)
                      category -> the category assigned to the artifact(s)
    config:         - a configuration dictionary containing at least
                      the following keys:
                      PATH_KB_DB        - the database path of KB
                      PATH_KB_DATA      - the data directory of KB
                      PATH_KB_HIST      - the history menu path of KB
    """
    initializer.init(config)

    if args["id"]:
        for i in args["id"]:
            delete_by_id(i, args["force"], config)

    elif args["title"]:
        delete_by_name(args["title"], args["category"], args["force"], config)


def delete_by_id(id: int, is_forced: bool, config: Dict[str, str]):
    """
    Edit the content of an artifact by id.

    Arguments:
    id:             - the ID (the one you see with kb list)
                      associated to the artifact to delete
    config:         - a configuration dictionary containing at least
                      the following keys:
                      PATH_KB_DB        - the database path of KB
                      PATH_KB_DATA      - the data directory of KB
                      PATH_KB_HIST      - the history menu path of KB
                      EDITOR            - the editor program to call
    """
    conn = db.create_connection(config["PATH_KB_DB"])
    artifact_id = history.get_artifact_id(config["PATH_KB_HIST"], id)
    artifact = db.get_artifact_by_id(conn, artifact_id)
    
    if not artifact:
        print("Error: Invalid artifact referenced")
        return

    if not is_forced: 
        confirm = ask_confirmation(artifact.title,artifact.category)
        if not artifact or not confirm:
            print("No artifact was removed")
            return

    db.delete_artifact_by_id(conn, artifact_id)

    category_path = Path(config["PATH_KB_DATA"], artifact.category)

    try:
        Path(category_path, artifact.title).unlink()
    except FileNotFoundError:
        pass

    if fs.count_files(category_path) == 0:
        fs.remove_directory(category_path)

    print("Artifact {category}/{title} removed!".format(
        category=artifact.category, title=artifact.title))


def delete_by_name(title: str, category: str, is_forced: bool, config: Dict[str, str]):
    """
    Edit the content of an artifact by name, that is title/category

    Arguments:
    title:          - the title assigned to the artifact to delete
    category:       - the category assigned to the artifact to delete
    config:         - a configuration dictionary containing at least
                      the following keys:
                      PATH_KB_DB        - the database path of KB
                      PATH_KB_DATA      - the data directory of KB
                      PATH_KB_HIST      - the history menu path of KB
                      EDITOR            - the editor program to call
    """
    conn = db.create_connection(config["PATH_KB_DB"])
    artifacts = db.get_artifacts_by_filter(conn, title=title,
                                           category=category,
                                           is_strict=True)
    if len(artifacts) == 1:
        artifact = artifacts.pop()

        if not is_forced:
            confirm = ask_confirmation(artifact.title,artifact.category)
            if not artifact or not confirm:
                print("No artifact was removed")
                return

        db.delete_artifact_by_id(conn, artifact.id)
        print("Artifact {}/{} removed!".format(artifact.category, artifact.title))
    elif len(artifacts) > 1:
        print(
            "There is more than one artifact with that title, please specify a category")
    else:
        print(
            "There is no artifact with that name, please specify a correct artifact name")


def ask_confirmation(title: str, category: str):
    """
    Ask confirmation for the deletion of an artifact

    Arguments:
    title:          - the title assigned to the artifact to delete
    category:       - the category assigned to the artifact to delete

    Returns:
    A boolean that is true if the user really wants to remove
    an artifact, i.e., "y" or "yes" have been typed at the prompt
    """
    answer = input(
        "Are you sure you want to delete {category}/{title}? [y/n]".format(
            category=category, title=title))

    return not (answer.lower() not in ["y","yes"])



# -*- encoding: utf-8 -*-
# kb v0.1.7
# A knowledge base organizer
# Copyright © 2020, gnc.
# See /LICENSE for licensing information.

"""
kb database module

:Copyright: © 2020, gnc.
:License: GPLv3 (see /LICENSE).
"""

__all__ = ()

import sqlite3
from pathlib import Path
from typing import List, Optional
import attr

from kb.entities.artifact import Artifact

DB_CREATE_QUERY = """CREATE TABLE IF NOT EXISTS artifacts (
                         id integer PRIMARY KEY,
                         title text NOT NULL,
                         category text NOT NULL,
                         path text NOT NULL,
                         tags text,
                         status text,
                         author text,
                         template text);
                     CREATE TABLE IF NOT EXISTS tags (
                         artifact_id integer,
                         tag text,
                         FOREIGN KEY(artifact_id)
                           REFERENCES artifacts(id)
                           ON DELETE CASCADE);
                  """


def create_connection(db_file: str):
    """
    Create a database connection to the SQLite database
    specified by db_file

    Arguments:
    db_file         -  database file

    Returns:
    An sqlite3 connection object or None in case of error
    """
    conn = None
    try:
        conn = sqlite3.connect(db_file)
        return conn
    except Exception as err:
        print("Connection Failed: {error}".format(error=err))

    return conn


def create_table(conn, create_table_sql: str) -> None:
    """
    Create a table from the create_table_sql statement
    or show an error message in case of failure

    Arguments:
    conn              - Connection object
    create_table_sql  - A 'CREATE TABLE' statement
    """
    try:
        cursor = conn.cursor()
        cursor.executescript(create_table_sql)
    except Exception as err:
        print("Table Creation Failed: {error}".format(error=err))


def create_kb_database(kb_db_path: str, schema_version: int) -> None:
    """
    Create an empty sqlite database for kb
    at the specified path or return an error
    if the creation is not possible

    Arguments:
    kb_db_path      - the path where the kb database will be stored
    """
    conn = create_connection(kb_db_path)

    if conn is not None:
        create_table(conn, DB_CREATE_QUERY)
        set_schema_version(conn, schema_version)
    else:
        print("Error! cannot create the database connection.")


def is_artifact_existing(conn, title: str, category: str) -> bool:
    """
    Checks if an artifact with the provided title and category
    already exists in the database

    Arguments:
    conn            - the sqlite3 connection object
    title           - the title of the artifact
    category        - the category of the artifact

    Returns:
    True if the artifact already exists in the database,
    otherwise False
    """
    cur = conn.cursor()
    cur.execute("""SELECT title,category
                   FROM artifacts
                   WHERE title=? AND category=?""", [title, category])

    result = cur.fetchone()
    return result is not None


def insert_artifact(conn, artifact: Artifact) -> None:
    """
    Inserts in the database the provided artifact

    Arguments:
    conn            - An sqlite connection object representing
                      the database on which data will be inserted
    artifact        - an artifact object

    Returns:
    Returns an error if there was a failure in the insertion operation
    """
    # Convert tags to a string with ';' separating tags
    tags_list = []
    if artifact.tags:
        tags_list = list(set(artifact.tags.split(';')))

    path = ""
    if artifact.path:
        path = artifact.path
    else:
        path = "{category}/{title}".format(
            category=artifact.category, title=artifact.title)

    cur = conn.cursor()
    if is_artifact_existing(conn, artifact.title, artifact.category):
        print("Error: the specified artifact already exists in kb!")
        print("Run `kb update -h` to understand how to update an artifact")
        return

    sql = '''INSERT INTO artifacts
             (title,category,path,tags,author,status,template)
             VALUES(?,?,?,?,?,?,?)'''
    args = (
        artifact.title,
        artifact.category,
        path,
        artifact.tags,
        artifact.author,
        artifact.status,
        artifact.template)

    cur.execute(sql, args)
    last_artifact_id = cur.lastrowid

    for tag in tags_list:
        sql = '''INSERT INTO tags
                 (artifact_id,tag)
                 VALUES(?,?)'''
        args = (last_artifact_id, tag)
        cur.execute(sql, args)

    conn.commit()


def insert_artifact_with_id(conn, artifact: Artifact, id: int) -> None:
    """
    Inserts in the database the provided artifact

    Arguments:
    conn            - An sqlite connection object representing
                      the database on which data will be inserted
    artifact        - an artifact object
    id              - the ID to use for the insertion of artifact
                      in the database

    Returns:
    Returns an error if there was a failure in the insertion operation
    """
    # Convert tags to a string with ';' separating tags
    tags_list = []
    if artifact.tags:
        tags_list = list(set(artifact.tags.split(';')))

    path = ""
    if artifact.path:
        path = artifact.path
    else:
        path = "{category}/{title}".format(
            category=artifact.category, title=artifact.title)

    cur = conn.cursor()
    if is_artifact_existing(conn, artifact.title, artifact.category):
        print("Error: the specified artifact already exists in kb!")
        print("Run `kb update -h` to understand how to update an artifact")
        return

    sql = '''INSERT INTO artifacts
             (id, title,category,path,tags,author,status,template)
             VALUES(?,?,?,?,?,?,?,?)'''
    args = (
        artifact.id,
        artifact.title,
        artifact.category,
        path,
        artifact.tags,
        artifact.author,
        artifact.status,
        artifact.template)

    cur.execute(sql, args)
    last_artifact_id = cur.lastrowid

    for tag in tags_list:
        sql = '''INSERT INTO tags
                 (artifact_id,tag)
                 VALUES(?,?)'''
        args = (last_artifact_id, tag)
        cur.execute(sql, args)

    conn.commit()


def delete_artifact_by_id(conn, artifact_id: int) -> None:
    """
    Deletes the artifact corresponding to the provided
    artifact ID

    Arguments:
    conn                - the sqlite3 connection object
    artifact_id         - the ID associated to the artifact to remove
    """
    cur = conn.cursor()
    sql_query = "DELETE FROM artifacts WHERE id = ?"

    cur.execute(sql_query, [artifact_id])
    conn.commit()


def delete_artifact_by_name(conn, title: str, category: str) -> None:
    """
    Deletes the artifact corresponding to the provided
    artifact name, that is title and category

    Arguments:
    conn                - the sqlite3 connection object
    artifact_id         - the ID associated to the artifact to remove
    """
    cur = conn.cursor()
    sql_query = "DELETE FROM artifacts WHERE title = ? AND category = ?"

    cur.execute(sql_query, [title, category])
    conn.commit()


def get_artifact_by_id(conn, artifact_id: int) -> Artifact:
    """
    Get the artifact corresponding to the provided
    artifact ID

    Arguments:
    conn                - the sqlite3 connection object
    artifact_id         - the ID associated to the artifact to retrieve
    """
    cur = conn.cursor()
    sql_query = "SELECT * FROM artifacts WHERE id = ?"
    cur.execute(sql_query, [artifact_id])

    result = cur.fetchone()
    if result:
        return Artifact(*result)


def get_artifacts_by_filter(
        conn,
        title: Optional[str] = None,
        category: Optional[str] = None,
        tags: Optional[List[str]] = None,
        author: Optional[str] = None,
        status: Optional[str] = None,
        is_strict: bool = False
) -> List[Artifact]:
    """
    Returns artifacts by applying a generic filter
    on title, category and tags.

    Arguments:
    conn                - the sqlite3 Connection object
    title               - the title string to match in the database
    category            - the category string to match in the database
    tags                - the tag list to match in the database
    author              - the author to match
    status              - the status to match
    is_strict           - if True, the pattern matching is
                          on exact strings

    Returns:
    a list of the artifacts matching the provided filter that
    has been found
    """
    artifact_id_list = list()

    if title is not None:
        artifacts_by_title = get_artifacts_by_title(
            conn, query_string=title, is_strict=is_strict)
        artifact_id_list.append({art.id for art in artifacts_by_title})
    if category is not None:
        artifacts_by_cat = get_artifacts_by_category(
            conn, query_string=category, is_strict=is_strict)
        artifact_id_list.append({art.id for art in artifacts_by_cat})
    if tags:
        artifacts_by_tags = get_artifacts_by_tags(
            conn, tags=tags, is_strict=is_strict)
        artifact_id_list.append({art.id for art in artifacts_by_tags})
    if author:
        artifacts_by_author = get_artifacts_by_author(
            conn, query_string=author, is_strict=is_strict)
        artifact_id_list.append({art.id for art in artifacts_by_author})
    if status:
        artifacts_by_status = get_artifacts_by_status(
            conn, query_string=status, is_strict=is_strict)
        artifact_id_list.append({art.id for art in artifacts_by_status})

    if len(artifact_id_list) == 0:
        return None

    # Perform an intersection on collected artifact IDs
    resulting_set = artifact_id_list.pop()
    for art_id in artifact_id_list:
        resulting_set.intersection_update(art_id)

    artifacts = []
    for result_id in resulting_set:
        artifacts.append(get_artifact_by_id(conn, result_id))

    return artifacts


def get_artifacts_by_title(
        conn,
        query_string: str = "",
        is_strict: bool = False
) -> List[Artifact]:
    """
    Returns artifacts matching the string to search,
    if the string is empty, it will retrieve all artifacts

    Arguments:
    conn                - the sqlite3 Connection object
    query_string        - the string to match in the database
    is_strict           - if True, the pattern matching is
                          on exact strings

    Returns:
    a list of the artifacts matching the provided string that
    have been found
    """
    if not is_strict:
        query_string = "%" + query_string + "%"
    cur = conn.cursor()
    sql_query = """
                SELECT *
                FROM artifacts
                WHERE title LIKE ?
                COLLATE NOCASE
                """
    cur.execute(sql_query, [query_string])
    artifacts = [Artifact(*row) for row in cur.fetchall()]
    return artifacts


def get_uniq_artifact_by_filter(
        conn,
        title: Optional[str] = None,
        category: Optional[str] = None,
        tags: Optional[List[str]] = None,
        author: Optional[str] = None,
        status: Optional[str] = None,
        is_strict: bool = False
) -> Artifact:
    """
    Get the artifact with the specified query string
    if the result is unique or None otherwise

    Arguments:
    conn                - the sqlite3 Connection object
    title               - the title string to match in the database
    category            - the category string to match in the database
    tags                - the tag list to match in the database
    author              - the author to match
    status              - the status to match
    is_strict           - if True, the pattern matching is
                          on exact strings
    Returns:
    This function can return either:
    - the Artifact object matching the string to search, if the
      result is unique
    - None if there is more than one artifact matching the
    query string or no artifact matched at all
    """
    artifacts = get_artifacts_by_filter(conn, title=title,
                                        category=category,
                                        tags=tags,
                                        author=author,
                                        status=status,
                                        is_strict=is_strict)
    if len(artifacts) == 1:
        return artifacts.pop()


def get_artifacts_by_status(
        conn,
        query_string: str = "",
        is_strict: bool = False
) -> List[Artifact]:
    """
    Returns artifacts matching the status string provided,
    if the string is empty, it will retrieve all artifacts

    Arguments:
    conn                - the sqlite3 Connection object
    query_string        - the status string to match in the database
    is_strict           - if True, the pattern matching is
                          on exact strings

    Returns:
    a list of the artifacts matching the provided status string that
    have been found
    """
    if not is_strict:
        query_string = "%" + query_string + "%"
    cur = conn.cursor()
    sql_query = """
                SELECT *
                FROM artifacts
                WHERE status LIKE ?
                COLLATE NOCASE
                """
    cur.execute(sql_query, [query_string])

    artifacts = [Artifact(*row) for row in cur.fetchall()]
    return artifacts


def get_artifacts_by_author(
        conn,
        query_string: str = "",
        is_strict: bool = False
) -> List[Artifact]:
    """
    Returns artifacts matching the string to search,
    if the string is empty, it will retrieve all artifacts

    Arguments:
    conn                - the sqlite3 Connection object
    query_string        - the author string to match in the database
    is_strict           - if True, the pattern matching is
                          on exact strings

    Returns:
    a set of the artifacts matching the provided author that
    have been found
    """
    if not is_strict:
        query_string = "%" + query_string + "%"
    cur = conn.cursor()
    sql_query = """
                SELECT *
                FROM artifacts
                WHERE author LIKE ?
                COLLATE NOCASE
                """
    cur.execute(sql_query, [query_string])
    artifacts = [Artifact(*row) for row in cur.fetchall()]
    return artifacts


def get_artifacts_by_category(
        conn,
        query_string: str = "",
        is_strict: bool = False
) -> List[Artifact]:
    """
    Returns artifacts of the category matching the string to search,
    if the string is empty, it will retrieve all artifacts

    Arguments:
    conn                - the sqlite3 Connection object
    query_string        - the string to match in the database
    is_strict           - if True, the pattern matching is
                          on exact strings

    Returns:
    a set of the artifacts matching the provided string that
    have been found
    """
    if not is_strict:
        query_string = "%" + query_string + "%"

    sql_query = """SELECT *
                   FROM artifacts
                   WHERE category LIKE ?
                   COLLATE NOCASE"""

    cur = conn.cursor()
    cur.execute(sql_query, [query_string])

    artifacts = [Artifact(*row) for row in cur.fetchall()]
    return artifacts


def get_artifacts_by_tags(
        conn,
        tags: List[str] = [],
        is_strict: bool = False
) -> List[Artifact]:
    """
    Returns artifacts matching the provided list of tags,
    if no tags are provided, it will retrieve all artifacts

    Arguments:
    conn                - the sqlite3 Connection object
    query_string        - the list of tags (strings) provided
    is_strict           - if True, the pattern matching is
                          on exact strings

    Returns:
    a list of the found artifacts matching the provided tags
    """
    rows = list()
    if not is_strict:
        tags = ["%" + tag + "%" for tag in tags]

    cur = conn.cursor()
    for tag in tags:
        sql_query = """
                    SELECT *
                    FROM artifacts INNER JOIN tags ON tags.artifact_id = artifacts.id
                    WHERE tag LIKE ?
                    COLLATE NOCASE
                    """
        cur.execute(sql_query, [tag])
        rows += cur.fetchall()

    # We discard the last two fields in result rows
    # since we are only interested in artifact data
    rows = [r[:-2] for r in list(set(rows))]

    artifacts = [Artifact(*row) for row in rows]
    return artifacts


def update_artifact_by_id(
        conn,
        artifact_id: int,
        artifact: Artifact
) -> None:
    """
    Update an artifact in the database

    Arguments:
    conn            - An sqlite connection object representing
                      the database on which data will be inserted
    artifact_id     - the ID associated to the artifact to update
    artifact        - an artifact object which contains the attributes
                      to update

    Returns:
    Returns an error if there was a failure in the update operation
    """
    current_artifact = get_artifact_by_id(conn, artifact_id)
    if not current_artifact:
        return None

    update_record = (artifact_id, artifact.title, artifact.category,
                     artifact.path, artifact.tags, artifact.status,
                     artifact.author, artifact.template)

    new_record = list()
    for i, elem in enumerate(attr.astuple(current_artifact)):
        new_record.append(update_record[i] or elem or None)

    delete_artifact_by_id(conn, artifact_id)
    updated_artifact = Artifact(*new_record)
    insert_artifact_with_id(conn, updated_artifact, artifact_id)


def get_schema_version(conn) -> int:
    """
    Check what is the version of the schema used by the
    database.

    Arguments:
    conn            - the database connection object

    Returns:
    An int number representing the version of the
    schema used by kb.
    """
    cur = conn.cursor()
    sql_query = """PRAGMA user_version;"""
    cur.execute(sql_query)
    return cur.fetchone()[0]


def set_schema_version(conn, version: int) -> int:
    """
    Check what is the version of the schema used by the
    database.

    Arguments:
    conn            - the database connection object
    version         - the version of the schema to be set (integer)

    Returns:
    An int number representing the version of the
    schema used by kb.
    """
    cur = conn.cursor()
    sql_query = "PRAGMA user_version = {v:d}".format(v=version)
    cur.execute(sql_query)
    conn.commit()


def is_schema_updated_to_version(conn, version: int) -> bool:
    """
    Check if the schema version of the database related to the
    connection object has the same version of the one provided.

    Arguments:
    conn            - the database connection object
    version         - the version to check

    Returns:
    A boolean that is True if the database schema version is
    equal to th passed version, False otherwise.
    """
    current_schema_version = get_schema_version(conn)
    return current_schema_version == version


def migrate_v0_to_v1(conn):
    """
    Migrates the database schema from v0 to v1.

    Arguments:
    conn            - the database connection object
    """
    cur = conn.cursor()
    sql_query = "ALTER TABLE artifacts ADD COLUMN template text"
    cur.execute(sql_query)
    conn.commit()
    set_schema_version(conn, 1)



# -*- encoding: utf-8 -*-
# kb v0.1.7
# A knowledge base organizer
# Copyright © 2020, gnc.
# See /LICENSE for licensing information.

"""
kb viewer module

:Copyright: © 2020, gnc.
:License: GPLv3 (see /LICENSE).
"""

import re
from typing import Dict
from kb.styler import set_fg, reset


def colorize_string(string, color):
    """
    This function applies the provided color to the specified string

    Arguments:
    msg   - The message to be colored
    color - The name of the color (or hex code), e.g., "red" or "#C0C0C0"

    Returns:
    A colored message
    """
    return set_fg(color) + string + reset()


def colorize_row(row, markers=None):
    """
    This function takes a string and a dictionary of markers
    (where key is regex and value is color) and
    the transforms the passed string (row) with colors.

    Arguments:
    row     - the string
    markers - a dictionary having as key strings
              representing the purpose of the formatting
              and having as values a list where
              the first element is a regex and the second
              element is the color value.
              Example:
              markers = {
                    "TITLE":    ["^#.*", "blue"]
                    "WARNINGS": ["^!.*", "yellow"]
              }

    Returns:
    A new message colored following the rules within the markers
    dictionary
    """
    colored_row = row
    for mark in markers:
        regex = re.compile(rf'{(markers[mark][0])}')
        color = markers[mark][1]

        match = regex.search(row)

        if match:
            colored_row = re.sub(
                regex, colorize_string(
                    match.group(0).replace(
                        "\\", "\\\\"), color), rf'{row}')
            row = colored_row

    return colored_row


def colorize_output(data, markers):
    """
    This function takes an input a list of strings, for example they
    can be taken from a file or any other source, processes them
    and returns a list of formatted colored strings ready to be
    visualized.

    Arguments:
    data    - A list of strings.
    markers - an object contains configured marks

    Returns:
    A new formatted list
    """
    if markers is None:
        return data
    colorized_output = list()
    for row in data:
        colorized_output.append(colorize_row(row, markers))
    return colorized_output


def view(filepath: str, markers: Dict[str, str], color: bool = True) -> None:
    """
    Visualize the specified file with applied markers
    if color is True.

    Arguments:
    filepath        - the file to visualize on stdout
    markers         - the markers dictionary to apply
    color           - a boolean, if True color is enabled
    """
    content = ""
    with open(filepath) as fname:
        content = fname.read()

    # Print on screen with proper markers
    lines = content.splitlines()
    if color:
        colored_lines = colorize_output(lines, markers)
        for line in colored_lines:
            print(line)
    else:
        for line in lines:
            print(line)



# -*- encoding: utf-8 -*-
# kb v0.1.7
# A knowledge base organizer
# Copyright © 2020, gnc.
# See /LICENSE for licensing information.

"""
kb erase command module

:Copyright: © 2020, gnc.
:License: GPLv3 (see /LICENSE).
"""

from typing import Dict
import kb.filesystem as fs


def erase(args: Dict[str, str], config: Dict[str, str]):
    """
    Erase the entire kb knowledge base (or only the database).

    Arguments:
    args:           - a dictionary containing the following fields:
                      db -> a boolean, if true, only the database
                        will be deleted
    config:         - a configuration dictionary containing at least
                      the following keys:
                      PATH_KB           - the main path of KB
                      PATH_KB_DB        - the database path of KB
                      PATH_KB_HIST      - the history menu path of KB
    """
    if args["db"]:
        answer = input(
            "Are you sure you want to erase the kb database ? [YES/NO]")
        if answer.lower() == "yes":
            try:
                fs.remove_file(config["PATH_KB_DB"])
                fs.remove_file(config["PATH_KB_HIST"])
                print("kb database deleted successfully!")
            except FileNotFoundError:
                pass
    else:
        answer = input(
            "Are you sure you want to erase the whole kb knowledge base ? [YES/NO]")
        if answer.lower() == "yes":
            try:
                fs.remove_directory(config["PATH_KB"])
                print("kb knowledge base deleted successfully!")
            except FileNotFoundError:
                pass



# -*- encoding: utf-8 -*-
# kb v0.1.7
# A knowledge base organizer
# Copyright © 2020, gnc.
# See /LICENSE for licensing information.

"""
kb printer style module for constants

:Copyright: © 2020, gnc.
:License: GPLv3 (see /LICENSE).
"""

from typing import List
import kb.styler as styler
from kb.entities.artifact import Artifact

ALT_BGROUND = styler.set_bg('#303030')
BOLD = styler.set_style('bold')
UND = styler.set_style('underlined')
RESET = styler.reset()



# -*- encoding: utf-8 -*-
# kb v0.1.7
# A knowledge base organizer
# Copyright © 2020, gnc.
# See /LICENSE for licensing information.

"""
kb styler module

:Copyright: © 2020, gnc.
:License: GPLv3 (see /LICENSE).

This is a gateway for the styler module used to styled text
"""
import colored


def set_bg(color: str) -> str:
    """
    Set background color.

    Arguments:
    color       - the color string to set, that
                  can be either a word (e.g., "green")
                  or an hex code (e.g., "#00AB00")
    Returns:
    A string representing the code to set the background color
    """
    return colored.bg(color)


def set_fg(color: str) -> str:
    """
    Set foreground color.

    Arguments:
    color       - the color string to set, that
                  can be either a word (e.g., "green")
                  or an hex code (e.g., "#00AB00")
    Returns:
    A string representing the code to set the foreground color
    """
    return colored.fg(color)


def set_style(style: str) -> str:
    """
    Set a specific text style

    Arguments:
    style       - a string representing the desired
                  style, examples:
                  "bold"
                  "underline"
    Returns:
    A string representing the code to set the desired style
    """
    return colored.attr(style)


def reset() -> str:
    """
    Reset applied style.

    Returns:
    A string representing the code to reset the style and colors to default
    """
    return colored.attr('reset')



# -*- encoding: utf-8 -*-
# kb v0.1.7
# A knowledge base organizer
# Copyright © 2020, gnc.
# See /LICENSE for licensing information.

"""
kb history module

:Copyright: © 2020, gnc.
:License: GPLv3 (see /LICENSE).
"""

from typing import List
from kb.entities.artifact import Artifact
import kb.db as db


def get_artifact_id(hist_file_path: str, list_id: int) -> int:
    """
    Get the database ID related to an artifact based
    on the list ID (ID shown by kb list).
    This function is based on the history file path
    that is used to retrieve the correspondence between
    list ID and database artifact ID

    Arguments:
    hist_file_path      - the path to the history file,
                          this is generally in
                          $HOME/.kb/recent.hist
    list_id             - the ID shown by kb list

    Returns:
    The database ID corresponding to the artifact or
    None in case of non-valid list ID
    """
    with open(hist_file_path, 'r') as hfile:
        for line in hfile:
            items = line.split(",")
            if items[0] == list_id:
                return items[1]
        return None


def write(hist_file_path: str, search_result: List) -> None:
    """
    Write the kb history file with the results of
    a search/list operation.

    Arguments:
    hist_file_path      - the path to the history file,
                          this is generally in
                          $HOME/.kb/recent.hist
    search_result       - the results returned from
                          a DB query
    """
    with open(hist_file_path, "w") as hfile:
        hfile.write('view_id,db_id\n')
        for view_id, result in enumerate(search_result):
            hfile.write("{},{}\n".format(view_id, result.id))


def get_artifact(conn, hist_file_path: str, list_id: int) -> Artifact:
    """
    Get an artifact based on the list ID (ID shown by kb list).
    This function is based on the history file path
    that is used to retrieve the correspondence between
    list ID and database artifact ID

    Arguments:
    hist_file_path      - the path to the history file,
                          this is by default in
                          $HOME/.kb/recent.hist
    list_id             - the ID shown by kb list

    Returns:
    The artifact corresponding to list_id shown by kb list
    None in case of non-valid list ID
    """
    artifact_id = get_artifact_id(
        hist_file_path, list_id)

    return db.get_artifact_by_id(conn, artifact_id)



# -*- encoding: utf-8 -*-
# kb v0.1.7
# A knowledge base organizer
# Copyright © 2020, gnc.
# See /LICENSE for licensing information.

"""
kb printer for search command module

:Copyright: © 2020, gnc.
:License: GPLv3 (see /LICENSE).
"""

from typing import List
from kb.printer.style import ALT_BGROUND, BOLD, UND, RESET
from kb.entities.artifact import Artifact


def generate_search_header(
        search_result: List[Artifact],
        color: bool = True
) -> str:
    """
    Generates kb query search results header.

    Arguments:
    search_result   - the list of Artifacts for generating
                      an adapted header
    color           - a boolean, True if color is enabled

    Returns:
    A string representing the header for the list of artifacts
    """

    if not search_result:
        return

    min_length = 20
    len_id = max(len(str(len(search_result) - 1)), 2)

    len_title = max(
        max([len(art.title) if art.title else 0 for art in search_result]), min_length)
    len_categ = max(max(
        [len(art.category) if art.category else 0 for art in search_result]), min_length)
    len_tags = max(
        max([len(art.tags) if art.tags else 0 for art in search_result]), min_length)

    header = "   [ {id} ]  {title} {category} {tags}".format(
        id="ID".rjust(len_id),
        title="Title".ljust(len_title),
        category="Category".ljust(len_categ),
        tags="Tags".ljust(len_tags))

    if color:
        return UND + BOLD + header + RESET
    return header


def generate_search_header_verbose(
        search_result: List[Artifact],
        color: bool = True
) -> str:
    """
    Generates kb query search results header in verbose mode.

    Arguments:
    search_result   - the list of Artifacts for generating
                      an adapted header
    color           - a boolean, True if color is enabled

    Returns:
    A string representing the header for the list of artifacts
    """
    if not search_result:
        return

    min_length = 20
    sec_min_length = 10
    len_id = max(len(str(len(search_result) - 1)), 2)

    len_title = max(
        max([len(art.title) if art.title else 0 for art in search_result]), min_length)
    len_categ = max(max(
        [len(art.category) if art.category else 0 for art in search_result]), min_length)
    len_tags = max(
        max([len(art.tags) if art.tags else 0 for art in search_result]), min_length)
    len_author = max(max(
        [len(art.author) if art.author else 0 for art in search_result]), sec_min_length)
    len_status = max(max(
        [len(art.status) if art.status else 0 for art in search_result]), sec_min_length)
    len_template = max(max(
        [len(art.template) if art.template else 0 for art in search_result]), sec_min_length)

    header = "   [ {id} ]  {title} {category} {tags} {author} {status} {template}".format(
        id="ID".rjust(len_id),
        title="Title".ljust(len_title),
        category="Category".ljust(len_categ),
        tags="Tags".ljust(len_tags),
        author="Author".ljust(len_author),
        status="Status".ljust(len_status),
        template="Template".ljust(len_template))

    if color:
        return UND + BOLD + header + RESET
    return header


def print_search_result(
        search_result: List[Artifact],
        color: bool = True
) -> None:
    """
    Print kb query search results

    Arguments:
    search_result   - the list of Artifacts to print
                      in the form of search result
    color           - a boolean, True if color is enabled
    """
    if not search_result:
        return

    print(generate_search_header(search_result, color=color))
    print()

    min_length = 20
    len_id = max(len(str(len(search_result) - 1)), 2)

    len_title = max(
        max([len(art.title) if art.title else 0 for art in search_result]), min_length)
    len_categ = max(max(
        [len(art.category) if art.category else 0 for art in search_result]), min_length)
    len_tags = max(
        max([len(art.tags) if art.tags else 0 for art in search_result]), min_length)

    # Print results
    for view_id, artifact in enumerate(search_result):
        tags = artifact.tags if artifact.tags else ""

        result_line = " - [ {id} ]  {title} {category} {tags}".format(
            id=str(view_id).rjust(len_id),
            title=artifact.title.ljust(len_title),
            category=artifact.category.ljust(len_categ),
            tags=tags.ljust(len_tags))

        if color and (view_id % 2 == 0):
            print(ALT_BGROUND + result_line + RESET)
        else:
            print(result_line)


def print_search_result_verbose(
        search_result: List[Artifact],
        color: bool = True
) -> None:
    """
    Print kb query search results in verbose mode.

    Arguments:
    search_result   - the list of Artifacts to print
                      in the form of search result
    color           - a boolean, True if color is enabled
    """
    if not search_result:
        return

    print(generate_search_header_verbose(search_result, color=color))
    print()

    min_length = 20
    sec_min_length = 10
    len_id = max(len(str(len(search_result) - 1)), 2)

    len_title = max(
        max([len(art.title) if art.title else 0 for art in search_result]), min_length)
    len_categ = max(max(
        [len(art.category) if art.category else 0 for art in search_result]), min_length)
    len_tags = max(
        max([len(art.tags) if art.tags else 0 for art in search_result]), min_length)
    len_author = max(max(
        [len(art.author) if art.author else 0 for art in search_result]), sec_min_length)
    len_status = max(max(
        [len(art.status) if art.status else 0 for art in search_result]), sec_min_length)
    len_template = max(max(
        [len(art.template) if art.template else 0 for art in search_result]), sec_min_length)

    # Print results
    for view_id, artifact in enumerate(search_result):
        tags = artifact.tags if artifact.tags else ""
        author = artifact.author if artifact.author else ""
        status = artifact.status if artifact.status else ""
        template = artifact.template if artifact.template else ""

        result_line = "   [ {id} ]  {title} {category} {tags} {author} {status} {template}".format(
            id=str(view_id).rjust(len_id),
            title=artifact.title.ljust(len_title),
            category=artifact.category.ljust(len_categ),
            tags=tags.ljust(len_tags),
            author=author.ljust(len_author),
            status=status.ljust(len_status),
            template=template.ljust(len_template))

        if color and (view_id % 2 == 0):
            print(ALT_BGROUND + result_line + RESET)
        else:
            print(result_line)


def print_search_result_full_mode(
        search_result: List[Artifact],
) -> None:
    """
    Print kb query search results in full-identifier mode, that is:
    category/title

    Arguments:
    search_result   - the list of Artifacts to print
                      in the form of search result
    """
    if not search_result:
        return
    # Print results
    for view_id, artifact in enumerate(search_result):

        result_line = "{category}/{title}".format(
            category=artifact.category,
            title=artifact.title)

        print(result_line)



# -*- encoding: utf-8 -*-
# kb v0.1.7
# A knowledge base organizer
# Copyright © 2020, gnc.
# See /LICENSE for licensing information.

"""
kb filesystem module

:Copyright: © 2020, gnc.
:License: GPLv3 (see /LICENSE).
"""

__all__ = ()

import os
import re
import shutil
import tempfile
from pathlib import Path
from typing import List


def list_files(directory: str) -> List[str]:
    """
    List all files contained in a directory recursively,
    similarly to the "find" UNIX command

    Args:
    directory       - a string representing the target directory

    Returns:
    A list of strings representing the path of files contained
    in the directory
    """

    # Get kbdir path
    dirpath = Path(directory)

    # Get list of files in the form: file1, dir1/file2, ...
    files = [str(f.relative_to(dirpath))
             for f in dirpath.rglob("*") if f.is_file()]
    return files


def list_dirs(directory: str) -> List[str]:
    """
    List all sub-directories contained in a directory

    Args:
    directory       - a string representing the path to a directory

    Returns:
    A list of strings representing the path of directories contained
    in the provided directory
    """
    # Get kbdir path
    dirpath = Path(directory)

    # Get list of files in the form: file1, dir1/file2, ...
    files = [str(f.relative_to(dirpath))
             for f in dirpath.rglob("*") if f.is_dir()]
    return files


def touch_file(filename: str):
    """
    Creates a new empty file, in the style of the UNIX
    touch program.

    Arguments:

    filename    - a path to a filename
    """
    Path(filename).touch()


def get_basename(filename: str) -> str:
    """
    Get basename for a file

    Arguments:
    filename    - a path to a filename

    Returns:
    The basename of the provided file
    """
    return Path(filename).name


def copy_file(source: str, dest: str) -> None:
    """
    Copies a file to the provided destination

    Arguments:
    source    - the path to the source file to copy
    dest      - the destination path of the copy
    """
    shutil.copy2(Path(source), Path(dest))


def remove_file(filename: str) -> None:
    """
    Removes a file from the filesystem

    Arguments:
    filename    - the file to remove from the kb directory
    """
    try:
        Path(filename).unlink()
    except FileNotFoundError:
        pass


def remove_directory(directory: str) -> None:
    """
    Removes a directory from the filesystem

    Arguments:
    directory    - the directory to remove from the kb system
    """
    shutil.rmtree(directory)


def create_directory(directory: str) -> None:
    """
    Create a directory if it does not exist.

    Arguments:
    directory    - the directory path to be created
    """
    os.makedirs(Path(directory), exist_ok=True)


def is_directory(path: str) -> bool:
    """
    Checks if the provided path is a directory.

    Arguments:
    path        - the path to check

    Returns:
    A boolean, if true, the path corresponds to a directory
    """
    return os.path.isdir(path)


def is_file(path: str) -> bool:
    """
    Checks if the provided path corresponds to a regular file.

    Arguments:
    path        - the path to check

    Returns:
    A boolean, if true, the path corresponds to a regular file
    """
    return os.path.isfile(path)


def count_files(directory: str) -> int:
    """
    Count the number of files in a directory

    Arguments:
    directory    - the directory where to count files

    Returns:
    the number of files contained in the directory
    """
    return len(list(Path(directory).iterdir()))


def move_file(source: str, dest: str) -> None:
    """
    Moves a file to the provided destination

    Arguments:
    source    - the path to the source file to copy
    dest      - the destination path of the copy
    """
    shutil.move(source, dest)


def get_temp_filepath() -> str:
    """
    Generates a temporary file path.

    Returns:
    A boolean, True if the file is of type text.
    """
    tmpfilename = None
    while tmpfilename is None:
        random_tmp_path = str(Path(tempfile.gettempdir(),
                                   os.urandom(24).hex()))
        if not os.path.isfile(random_tmp_path):
            tmpfilename = random_tmp_path
    return tmpfilename


def is_text_file(filename: str) -> bool:
    """
    Determines if a file is textual (that can be
    nicely viewed in a text editor) or belonging
    to other types.

    Arguments:
    filename        - the file name/path to check

    Returns:
    A boolean, True if the file is of type text.
    """
    txt_extensions = ("", ".conf", ".ini", ".txt",
                      ".md", ".rst", ".ascii", ".org", ".tex")

    file_ext = os.path.splitext(filename)[1]

    return file_ext in txt_extensions


def get_filename_parts_wo_prefix(
        filename: str,
        prefix_to_remove: str) -> List[str]:
    """
    Get filename parts without the provided prefix.
    E.g., if the filename is "/path/to/data/dir1/file2.txt"
    and the prefix to remove is "/path/to/data" then the
    returned will be a tuple containing ("dir1","file2.txt")

    Arguments:
    filename          - a string or path provided by pathlib
    prefix_to_remove  - a string or path provided by pathlib that
                        will be removed from the filename

    Returns:
    The provided filename without the provided prefix
    """
    prefix_path = Path(prefix_to_remove)
    file_path = Path(filename)

    try:
        return file_path.relative_to(prefix_path).parts
    except ValueError:
        file_path.parts


def grep_in_files(
        filelist: str,
        regex: str,
        case_insensitive: bool = False
) -> List[str]:
    """
    Grep recursively through a file list by trying to match
    a regex with the content of all the files.
    This is equivalent to:
    grep -nr 'regex' file1 file2 ...

    Arguments:
    filelist     - the file list where to match the regex
    regex        - the regex to match

    Returns
    A list of tuples containing (filepath, line number, matched string)
    """
    if case_insensitive:
        pattern = re.compile(regex, re.IGNORECASE)
    else:
        pattern = re.compile(regex)

    matches = list()
    for fname in filelist:
        try:
            with open(fname) as handle:
                for i, line in enumerate(handle):
                    match = pattern.search(line)
                    linenumber = i + 1
                    if match:
                        matches.append((fname, linenumber, line.strip()))
                        # !!!TODO: This can be of inspiration for later result show
                        # print("%s:%s: %s" % (filepath, lineno,
                        #     line.strip().replace(mo.group(), "\033[92m%s\033[0m"%
                        #         mo.group())))
        except UnicodeDecodeError:
            # In this case the file is binary,
            # so we don't search through binary files
            continue
    return matches


def grep_in_files_uniq(
        filelist: str,
        regex: str,
        case_insensitive=False
) -> List[str]:
    """
    Grep recursively through a list of files by trying to match
    a regex with the content of all the found files.
    Note that it will return only the set of filenames matched without
    the content.
    This is equivalent to:
    grep -lnr 'regex' file1 file2 ...

    Arguments:
    filelist     - the file list where to match the regex
    regex        - the regex to match

    Returns
    A list of file paths matching the regex.
    """
    if case_insensitive:
        pattern = re.compile(regex, re.IGNORECASE)
    else:
        pattern = re.compile(regex)

    matches = list()
    for fname in filelist:
        try:
            with open(fname) as handle:
                for line in handle:
                    match = pattern.search(line)
                    if match:
                        matches.append(fname)
        except UnicodeDecodeError:
            # In this case the file is binary,
            # so we don't search through binary files
            continue
    return list(set(matches))



# -*- encoding: utf-8 -*-
# kb v0.1.7
# A knowledge base organizer
# Copyright © 2020, gnc.
# See /LICENSE for licensing information.

"""
kb main module

:Copyright: © 2020, gnc.
:License: GPLv3 (see /LICENSE).
"""

__all__ = ()

import sys
from kb.cl_parser import parse_args

from kb.commands.add import add
from kb.commands.search import search
from kb.commands.edit import edit
from kb.commands.update import update
from kb.commands.delete import delete
from kb.commands.template import template
from kb.commands.view import view
from kb.commands.grep import grep
from kb.commands.erase import erase
from kb.commands.ingest import ingest
from kb.commands.export import export
from kb.commands.sync import sync

from kb.config import DEFAULT_CONFIG


COMMANDS = {
    'add': add,
    'delete': delete,
    'edit': edit,
    'update': update,
    'list': search,
    'view': view,
    'grep': grep,
    'erase': erase,
    'import': ingest,
    'export': export,
    'template': template,
    'sync': sync,
}


def dispatch(function, *args, **kwargs):
    """
    Dispatch command line action to proper
    kb function
    """
    return COMMANDS[function](*args, **kwargs)


def main():
    """Main routine of kb."""
    args = parse_args(sys.argv[1:])

    cmd = args.command
    cmd_params = vars(args)

    dispatch(cmd, cmd_params, config=DEFAULT_CONFIG)



# -*- encoding: utf-8 -*-
# kb v0.1.7
# A knowledge base organizer
# Copyright © 2020, gnc.
# See /LICENSE for licensing information.

"""
kb initializer module

:Copyright: © 2020, gnc.
:License: GPLv3 (see /LICENSE).
"""

import os
from pathlib import Path
import toml
import kb.db as db
import kb.filesystem as fs
import kb.config as conf


def init(config):
    """
    Initialize kb with the provided configuration.

    Arguments:
    config  - a dictionary containing the following keys:
                PATH_KB                 - the path to kb
                                            (~/.kb by default)
                PATH_KB_DB              - the path to kb database
                                            (~/.kb/kb.db by default)
                PATH_KB_DATA            - the path to kb data
                                            (~/.kb/data/ by default)
                PATH_KB_DEFAULT_TEMPLATE - the path to kb markers
                                            (~/.kb/templates/default by default)
    """
    if not is_initialized(config):
        create_kb_files(config)


def create_kb_files(config):
    """
    Create the kb files and infrastructure

    Arguments:
    config  - a dictionary containing the following keys:
                PATH_KB                   - the path to kb
                                            (~/.kb by default)
                PATH_KB_DB                - the path to kb database
                                            (~/.kb/kb.db by default)
                PATH_KB_DATA              - the path to kb data
                                            (~/.kb/data/ by default)
                INITIAL_CATEGORIES        - a list containing the initial
                                            categories contained within kb
                PATH_KB_TEMPLATES         - the path to kb templates
                                            (~/.kb/templates/ by default)
                DB_SCHEMA_VERSION         - the database schema version
    """
    # Get paths for kb from configuration
    kb_path = config["PATH_KB"]
    db_path = config["PATH_KB_DB"]
    data_path = config["PATH_KB_DATA"]
    initial_categs = config["INITIAL_CATEGORIES"]
    templates_path = config["PATH_KB_TEMPLATES"]
    schema_version = config["DB_SCHEMA_VERSION"]
    default_template_path = str(Path(templates_path) / "default")

    # Create main kb
    fs.create_directory(kb_path)

    # Create kb database
    if not os.path.exists(db_path):
        db.create_kb_database(db_path, schema_version)

    # Check schema version
    conn = db.create_connection(db_path)
    current_schema_version = db.get_schema_version(conn)

    if current_schema_version == 0:
        db.migrate_v0_to_v1(conn)

    # Create "data" directory
    fs.create_directory(data_path)

    # Create "templates" directory
    fs.create_directory(templates_path)

    # Create kb initial categories directories
    for category in initial_categs:
        category_path = Path(data_path, category)
        fs.create_directory(category_path)

    # Create markers file
    with open(default_template_path, 'w') as cfg:
        cfg.write(toml.dumps(conf.DEFAULT_TEMPLATE))


def is_initialized(config) -> bool:
    """
    Check if kb is correctly initialized,
    ensure that:
    1 - the .kb directory exists
    2 - the kb database exists
    3 - the kb data directory exists

    Arguments:
    config  - a dictionary containing the following keys:
                PATH_KB         - the path to kb
                                    (~/.kb by default)
                PATH_KB_DB      - the path to kb
                                    (~/.kb/kb.db by default)
                PATH_KB_DATA    - the path to kb
                                    (~/.kb/data/ by default)
    Returns:
    True is kb is correctly initialized, False otherwise
    """
    kb_path = config["PATH_KB"]
    db_path = config["PATH_KB_DB"]
    data_path = config["PATH_KB_DATA"]
    templates_path = config["PATH_KB_TEMPLATES"]

    for path in [kb_path, db_path, data_path, templates_path]:
        if not os.path.exists(path):
            return False

    conn = db.create_connection(db_path)
    return db.is_schema_updated_to_version(conn, config["DB_SCHEMA_VERSION"])



# -*- encoding: utf-8 -*-
# kb v0.1.7
# A knowledge base organizer
# Copyright © 2020, gnc.
# See /LICENSE for licensing information.

"""
kb opener module

:Copyright: © 2020, gnc.
:License: GPLv3 (see /LICENSE).
"""

import os
import subprocess
import platform


def open_non_text_file(filename):
    """
    Open a non-text file
    """
    if platform.system() == 'Darwin':
        subprocess.Popen(['open', filename])
    elif platform.system() == 'Windows':
        os.startfile(filename)
    else:
        subprocess.Popen(['xdg-open', filename])



# -*- encoding: utf-8 -*-
# kb v0.1.7
# A knowledge base organizer
# Copyright © 2020, gnc.
# See /LICENSE for licensing information.

"""
kb add command module

:Copyright: © 2020, gnc.
:License: GPLv3 (see /LICENSE).
"""

import shlex
import sys
from pathlib import Path
from subprocess import call
from typing import Dict
import kb.db as db
import kb.initializer as initializer
import kb.filesystem as fs
from kb.entities.artifact import Artifact


def add(args: Dict[str, str], config: Dict[str, str]):
    """
    Adds a list of artifacts to the knowledge base of kb.

    Arguments:
    args:           - a dictionary containing the following fields:
                      file -> a list of files to add to kb
                      title -> the title assigned to the artifact(s)
                      category -> the category assigned to the artifact(s)
                      tags -> the tags assigned to the artifact(s)
                      author -> the author to assign to the artifact
                      status -> the status to assign to the artifact
    config:         - a configuration dictionary containing at least
                      the following keys:
                      PATH_KB_DB        - the database path of KB
                      PATH_KB_DATA      - the data directory of KB
                      EDITOR            - the editor program to call
    """
    # Check if the add command has proper arguments/options
    is_valid_add = args["file"] or args["title"]
    if not is_valid_add:
        print("Please, either specify a file or a title for the new artifact")
        sys.exit(1)

    # Check initialization
    initializer.init(config)

    conn = db.create_connection(config["PATH_KB_DB"])
    if args["file"]:
        for fname in args["file"]:
            if fs.is_directory(fname):
                continue
            add_file_to_kb(conn, args, config, fname)
    else:
        # Get title for the new artifact
        title = args["title"]

        # Assign a "default" category if not provided
        category = args["category"] or "default"

        # Create "category" directory if it does not exist
        category_path = Path(config["PATH_KB_DATA"], category)
        category_path.mkdir(parents=True, exist_ok=True)

        is_stdin_empty = sys.stdin.isatty()

        if not db.is_artifact_existing(conn, title, category):
            # If a file is provided, copy the file to kb directory
            # otherwise open up the editor and create some content
            artifact_path = str(Path(category_path, title))
            if args["body"]:
                with open(artifact_path, "w+") as art_file:
                    body = args["body"].replace("\\n", "\n")
                    art_file.write(body)
            elif not is_stdin_empty:
                with open(artifact_path, "w+") as art_file:
                    body_lines = sys.stdin.readlines()
                    art_file.writelines(body_lines)
            else:
                shell_cmd = shlex.split(
                    config["EDITOR"]) + [artifact_path]
                call(shell_cmd)

        new_artifact = Artifact(
            id=None, title=title, category=category,
            path="{category}/{title}".format(category=category, title=title),
            tags=args["tags"],
            status=args["status"], author=args["author"], template=args["template"])
        db.insert_artifact(conn, new_artifact)


def validate(args):
    """
    Validate arguments for the add command

    Arguments:
    args        - the dictionary of arguments
                  passed to the add command

    Returns:
    A boolean, True if the add command is valid
    """
    return bool(args["file"] or args["title"])


def add_file_to_kb(
        conn,
        args: Dict[str, str],
        config: Dict[str, str],
        fname: str
) -> None:
    """
    Adds a file to the kb knowledge base.

    Arguments:
    conn        -   the connection to the database object
    args        -   the args dictionary passed to the add command,
                    it must contain at least the following keys:
                        title, category, tags, status, author
    config      -   the configuration dictionary that must contain
                    at least the following key:
                    PATH_KB_DATA, the path to where artifact are stored
    fname       -   the path of the file to add to kb
    """
    title = args["title"] or fs.get_basename(fname)
    category = args["category"] or "default"
    template = args["template"] or "default"

    category_path = Path(config["PATH_KB_DATA"], category)
    category_path.mkdir(parents=True, exist_ok=True)

    try:
        fs.copy_file(fname, Path(category_path, title))
    except FileNotFoundError:
        print("Error: The specified file does not exist!".format(fname=fname))
        sys.exit(1)

    if not db.is_artifact_existing(conn, title, category):
        fs.copy_file(fname, Path(category_path, title))

    new_artifact = Artifact(
        id=None,
        title=title, category=category,
        path="{category}/{title}".format(category=category, title=title),
        tags=args["tags"],
        status=args["status"], author=args["author"], template=template)
    db.insert_artifact(conn, new_artifact)



# -*- encoding: utf-8 -*-
# kb v0.1.7
# A knowledge base organizer
# Copyright © 2020, gnc.
# See /LICENSE for licensing information.

"""
kb printer for grep command module

:Copyright: © 2020, gnc.
:License: GPLv3 (see /LICENSE).
"""

import pathlib
from typing import List
from kb.config import DEFAULT_CONFIG as config
import kb.filesystem as fs
from kb.printer.style import ALT_BGROUND, BOLD, UND, RESET
from kb.entities.artifact import Artifact


def generate_grep_header(
        grep_result: List[Artifact],
        hits_list: List[int],
        color: bool = True
) -> None:
    """
    Generates kb grep results header.

    Arguments:
    grep_result     - the list of Artifacts for generating
                      an adapted header
    hits_list       - a list ordered as the artifacts
                      representing the number of grep
                      matches for that artifact
    color           - a boolean, True if color is enabled

    Returns:
    A string representing the header for the list of artifacts
    """
    if not grep_result:
        return

    min_length = 20
    len_id = max(len(str(len(grep_result) - 1)), 2)

    len_title = max(
        max([len(art.title) if art.title else 0 for art in grep_result]), min_length)
    len_categ = max(max(
        [len(art.category) if art.category else 0 for art in grep_result]), min_length)
    len_tags = max(
        max([len(art.tags) if art.tags else 0 for art in grep_result]), min_length)

    len_hits = max(len(str(max(hits_list))), 4)

    header = "   [ {id} ]  {title} {category} {hits} {tags}".format(
        id="ID".rjust(len_id),
        title="Title".ljust(len_title),
        category="Category".ljust(len_categ),
        hits="Hits".ljust(len_hits),
        tags="Tags".ljust(len_tags))

    if color:
        return UND + BOLD + header + RESET
    return header


def generate_grep_header_verbose(
        grep_result: List[Artifact],
        hits_list: List[int],
        color: bool = True
) -> None:
    """
    Generates kb grep results header in verbose mode.

    Arguments:
    grep_result     - the list of Artifacts for generating
                      an adapted header
    hits_list       - a list ordered as the artifacts
                      representing the number of grep
                      matches for that artifact
    color           - a boolean, True if color is enabled

    Returns:
    A string representing the header for the list of artifacts
    """

    if not grep_result:
        return

    min_length = 20
    sec_min_length = 10
    len_id = max(len(str(len(grep_result) - 1)), 2)

    len_title = max(
        max([len(art.title) if art.title else 0 for art in grep_result]), min_length)
    len_categ = max(max(
        [len(art.category) if art.category else 0 for art in grep_result]), min_length)
    len_tags = max(
        max([len(art.tags) if art.tags else 0 for art in grep_result]), min_length)
    len_author = max(max(
        [len(art.author) if art.author else 0 for art in grep_result]), sec_min_length)
    len_status = max(max(
        [len(art.status) if art.status else 0 for art in grep_result]), sec_min_length)

    len_hits = max(len(str(max(hits_list))), 4)

    header = "   [ {id} ]  {title} {category} {hits} {tags} {author} {status}".format(
        id="ID".rjust(len_id),
        title="Title".ljust(len_title),
        category="Category".ljust(len_categ),
        hits="Hits".ljust(len_hits),
        tags="Tags".ljust(len_tags),
        author="Author".ljust(len_author),
        status="Status".ljust(len_status))

    if color:
        return UND + BOLD + header + RESET
    return header


def print_grep_result(
        grep_result: List[Artifact],
        hits_list: List[int],
        color: bool = True
) -> None:
    """
    Print kb query grep results.

    Arguments:
    grep_result     - the list of Artifacts to print
                      in the form of grep result
    hits_list       - a list ordered as the artifacts
                      representing the number of grep
                      matches for that artifact
    color           - a boolean, if True, color is enabled
    """
    if not grep_result:
        return

    print(generate_grep_header(grep_result, hits_list, color=color))
    print()

    min_length = 20

    len_id = max(len(str(len(grep_result) - 1)), 2)

    len_title = max(
        max([len(art.title) if art.title else 0 for art in grep_result]), min_length)
    len_categ = max(max(
        [len(art.category) if art.category else 0 for art in grep_result]), min_length)
    len_tags = max(
        max([len(art.tags) if art.tags else 0 for art in grep_result]), min_length)

    len_hits = max(len(str(max(hits_list))), 4)

    # Print results
    for view_id, artifact in enumerate(grep_result):
        tags = artifact.tags if artifact.tags else ""

        hits_id = view_id
        hits = str(hits_list[hits_id])

        result_line = "   [ {id} ]  {title} {category} {hits} {tags}".format(
            id=str(view_id).rjust(len_id),
            title=artifact.title.ljust(len_title),
            category=artifact.category.ljust(len_categ),
            hits=hits.ljust(len_hits),
            tags=tags.ljust(len_tags))

        if color and (view_id % 2 == 0):
            print(ALT_BGROUND + result_line + RESET)
        else:
            print(result_line)


def print_grep_result_verbose(
        grep_result: List[Artifact],
        hits_list: List[int],
        color: bool = True
) -> None:
    """
    Print more verbose kb query grep results.

    Arguments:
    grep_result     - the list of Artifacts to print
                      sorted by number of hits
    hits_list       - a list ordered as the grep_result artifacts
                      (i.e., by number of hits) representing the number
                      of grep matches for that artifact
    color           - a boolean, if True, color is enabled
    """
    if not grep_result:
        return

    print(generate_grep_header(grep_result, hits_list, color=color))
    print()

    min_length = 20
    sec_min_length = 10

    len_id = max(len(str(len(grep_result) - 1)), 2)

    len_title = max(
        max([len(art.title) if art.title else 0 for art in grep_result]), min_length)
    len_categ = max(max(
        [len(art.category) if art.category else 0 for art in grep_result]), min_length)
    len_tags = max(
        max([len(art.tags) if art.tags else 0 for art in grep_result]), min_length)
    len_author = max(max(
        [len(art.author) if art.author else 0 for art in grep_result]), sec_min_length)
    len_status = max(max(
        [len(art.status) if art.status else 0 for art in grep_result]), sec_min_length)

    len_hits = max(len(str(max(hits_list))), 4)

    # Print results
    for view_id, artifact in enumerate(grep_result):
        tags = artifact.tags if artifact.tags else ""

        author = artifact.author if artifact.author else ""
        status = artifact.status if artifact.status else ""

        hits_id = view_id
        hits = str(hits_list[hits_id])

        result_line = "   [ {id} ]  {title} {category} {hits} {tags} {author} {status}".format(
            id=str(view_id).rjust(len_id),
            title=artifact.title.ljust(len_title),
            category=artifact.category.ljust(len_categ),
            hits=hits.ljust(len_hits),
            tags=tags.ljust(len_tags),
            author=author.ljust(len_author),
            status=status.ljust(len_status))

        if color and (view_id % 2 == 0):
            print(ALT_BGROUND + result_line + RESET)
        else:
            print(result_line)

# This function still has to be implemented, this is just a placeholder


def print_grep_matches(grep_matches, color=True):
    """
    Print text associated to grep matches.

    Arguments:
    grep_matches    - the list of Artifacts to print
                      in the form of grep matches
    color           - a boolean, if True, color is enabled
    """

    for view_id, match in enumerate(grep_matches):
        path = "/".join(
            fs.get_filename_parts_wo_prefix(
                match[0], config["PATH_KB_DATA"]))
        line_number = match[1]
        matched_text = match[2]

        if color:
            path = BOLD + str(path) + RESET
            line_number = BOLD + str(line_number) + RESET

        result_line = "{path}:{line_number}:{matched_text}".format(
            path=path,
            line_number=line_number,
            matched_text=matched_text)
        print(result_line)



# -*- encoding: utf-8 -*-
# kb v0.1.7
# A knowledge base organizer
# Copyright © 2020, gnc.
# See /LICENSE for licensing information.

"""
kb printer for template command module

:Copyright: © 2020, gnc.
:License: GPLv3 (see /LICENSE).
"""

from typing import List
from kb.printer.style import ALT_BGROUND, BOLD, UND, RESET
from kb.entities.artifact import Artifact


def generate_template_search_header(
        color: bool = True
) -> str:
    """
    Generates kb query template results header.

    Arguments:
    color           - a boolean, True if color is enabled

    Returns:
    A string representing the header for the list of templates
    """
    header = "Templates"

    if color:
        return UND + BOLD + header + RESET
    return header


def print_template_search_result(
        template_search_result: List[str],
        color: bool = True
) -> None:
    """
    Print kb template search results.

    Arguments:
    template_search_result  - the list of templates
    color                   - a boolean, True if color is enabled
    """

    print(generate_template_search_header(color=color))
    print()

    len_template_name = max([len(template)
                            for template in template_search_result])

    # Print template search results
    for view_id, template in enumerate(template_search_result):
        result_line = " - {template}".format(
            template=template.ljust(len_template_name))

        if color and (view_id % 2 == 0):
            print(ALT_BGROUND + result_line + RESET)
        else:
            print(result_line)



# -*- encoding: utf-8 -*-
# kb v0.1.7
# A knowledge base organizer
# Copyright © 2020, gnc.
# See /LICENSE for licensing information.

"""
kb edit command module

:Copyright: © 2020, gnc.
:License: GPLv3 (see /LICENSE).
"""

import shlex
from subprocess import call
from typing import Dict
from pathlib import Path
import kb.db as db
import kb.initializer as initializer
import kb.history as history
import kb.filesystem as fs
from kb.entities.artifact import Artifact


def update(args: Dict[str, str], config: Dict[str, str]):
    """
    Update artifact properties within the knowledge base of kb.

    Arguments:
    args:           - a dictionary containing the following fields:
                      id -> a list of IDs (the ones you see with kb list)
                        associated to the artifact to update
                      title -> the title to be assigned to the artifact
                        to update
                      category -> the category to be assigned to the
                        artifact to update
                      tags -> the tags to be assigned to the artifact
                        to update
                      author -> the author to be assigned to the artifact
                        to update
                      status -> the status to be assigned to the artifact
                        to update
                      template -> the template to be assigned to the artifact
                        to update
                      edit_content -> a boolean, if True -> also open the
                        artifact to edit the content
    config:         - a configuration dictionary containing at least
                      the following keys:
                      PATH_KB_DB        - the database path of KB
                      PATH_KB_DATA      - the data directory of KB
                      PATH_KB_HIST      - the history menu path of KB
                      EDITOR            - the editor program to call
    """
    initializer.init(config)

    conn = db.create_connection(config["PATH_KB_DB"])

    # if an ID is specified, load artifact with that ID
    if args["id"]:
        old_artifact = history.get_artifact(conn,
                                            config["PATH_KB_HIST"], args["id"])
        if not old_artifact:
            print("The artifact you are trying to update does not exist! "
                  "Please insert a valid ID...")
            return None

        updated_artifact = Artifact(
            id=None,
            title=args["title"],
            category=args["category"],
            tags=args["tags"],
            author=args["author"],
            status=args["status"],
            template=args["template"])

        db.update_artifact_by_id(conn, old_artifact.id, updated_artifact)
        # If either title or category has been changed, we must move the file
        if args["category"] or args["title"]:
            old_category_path = Path(
                config["PATH_KB_DATA"],
                old_artifact.category)
            new_category_path = Path(
                config["PATH_KB_DATA"],
                args["category"] or old_artifact.category)
            fs.create_directory(new_category_path)

            fs.move_file(Path(old_category_path, old_artifact.title), Path(
                new_category_path, args["title"] or old_artifact.title))
    # else if a title is specified
    elif args["title"]:
        artifact = db.get_uniq_artifact_by_filter(conn, title=args["title"],
                                                  category=args["category"],
                                                  author=args["author"],
                                                  status=args["status"],
                                                  is_strict=True)

        if artifact:
            category_path = Path(config["PATH_KB_DATA"], artifact.category)
        else:
            print(
                "There is none or more than one artifact with that title, please specify a category")

    if args["edit_content"] or args["body"]:
        if args["title"]:
            artifact_path = str(Path(category_path, artifact.title))
            shell_cmd = shlex.split(config["EDITOR"]) + [artifact_path]
        elif args["id"]:
            artifact_path = str(Path(config["PATH_KB_DATA"])
                                / old_artifact.category
                                / old_artifact.title)
            shell_cmd = shlex.split(config["EDITOR"]) + [artifact_path]

        if args["body"]:
            args["body"] = args["body"].replace("\\n", "\n")
            with open(artifact_path, 'w') as art_file:
                art_file.write(args["body"])
        else:
            call(shell_cmd)



# -*- encoding: utf-8 -*-
# kb v0.1.7
# A knowledge base organizer
# Copyright © 2020, gnc.
# See /LICENSE for licensing information.

"""
kb view command module

:Copyright: © 2020, gnc.
:License: GPLv3 (see /LICENSE).
"""

import os
import platform
import sys
import shlex
import tempfile
from subprocess import call
from pathlib import Path
from typing import Dict
import kb.db as db
import kb.filesystem as fs
import kb.history as history
import kb.initializer as initializer
import kb.opener as opener
import kb.viewer as viewer
from kb.config import get_markers
from kb.entities.artifact import Artifact


def view(args: Dict[str, str], config: Dict[str, str]):
    """
    View an artifact contained in the knowledge base of kb.

    Arguments:
    args:           - a dictionary containing the following fields:
                      id -> the IDs (the one you see with kb list)
                        associated to the artifact to view
                      title -> the title of the artifact to view
                      category -> the category of the artifact to view
                      editor -> a boolean, if True the file will
                        be opened in a text editor as a temporary file
                        hence the original will not be affected
    config:         - a configuration dictionary containing at least
                      the following keys:
                      PATH_KB_DB                 - the database path of KB
                      PATH_KB_DATA               - the data directory of KB
                      PATH_KB_HIST               - the history menu path of KB
                      PATH_KB_DEFAULT_TEMPLATE   - the file associated to the markers
                      EDITOR                     - the editor program to call
    """
    # Check initialization
    initializer.init(config)

    color_mode = not args["no_color"]
    if args["id"]:
        view_by_id(args["id"], config, args["editor"], color_mode)
    elif args["title"]:
        view_by_name(
            args["title"],
            args["category"],
            config,
            args["editor"],
            color_mode)
    elif args["nameid"]:
        if args["nameid"].isdigit():
            view_by_id(args["nameid"], config, args["editor"], color_mode)
        else:
            view_by_name(
                args["nameid"],
                args["category"],
                config,
                args["editor"],
                color_mode)


def view_by_id(id: int,
               config: Dict[str,
                            str],
               open_editor: bool,
               color_mode: bool):
    """
    View the content of an artifact by id.

    Arguments:
    id:             - the ID (the one you see with kb list)
                      associated to the artifact we want to edit
    config:         - a configuration dictionary containing at least
                      the following keys:
                      PATH_KB_DB        - the database path of KB
                      PATH_KB_DATA      - the data directory of KB
                      PATH_KB_HIST      - the history menu path of KB
                      EDITOR            - the editor program to call
    open_editor     - a boolean, if True it will open the artifact as
                      a temporary copy in editor
    color_mode      - a boolean, if True the colors on screen will be
                      enabled when printed on stdout
    """
    conn = db.create_connection(config["PATH_KB_DB"])
    artifact_id = history.get_artifact_id(
        config["PATH_KB_HIST"], id)

    artifact = db.get_artifact_by_id(conn, artifact_id)

    if not artifact:
        sys.exit(1)

    category_path = Path(config["PATH_KB_DATA"], artifact.category)
    artifact_path = Path(category_path, artifact.title)

    if open_editor:
        tmpfname = fs.get_temp_filepath()
        fs.copy_file(artifact_path, tmpfname)

        shell_cmd = shlex.split(config["EDITOR"]) + [tmpfname]
        call(shell_cmd)
        fs.remove_file(tmpfname)

        sys.exit(0)

    # View File
    if fs.is_text_file(artifact_path):
        markers = get_template(artifact, config)
        viewer.view(artifact_path, markers, color=color_mode)
    else:
        opener.open_non_text_file(artifact_path)


def view_by_name(title: str,
                 category: str,
                 config: Dict[str,
                              str],
                 open_editor: bool,
                 color_mode: bool):
    """
    View the content of an artifact by name, that is title/category

    Arguments:
    title:          - the title assigned to the artifact(s)
    category:       - the category assigned to the artifact(s)
    config:         - a configuration dictionary containing at least
                      the following keys:
                      PATH_KB_DB        - the database path of KB
                      PATH_KB_DATA      - the data directory of KB
                      PATH_KB_HIST      - the history menu path of KB
                      EDITOR            - the editor program to call
    open_editor     - a boolean, if True it will open the artifact as
                      a temporary copy in editor
    color_mode      - a boolean, if True the colors on screen will be
                      enabled when printed on stdout
    """
    conn = db.create_connection(config["PATH_KB_DB"])
    artifacts = db.get_artifacts_by_filter(conn, title=title,
                                           category=category,
                                           is_strict=True)
    if len(artifacts) == 1:
        artifact = artifacts.pop()
        category_path = Path(config["PATH_KB_DATA"], artifact.category)
        artifact_path = Path(category_path, artifact.title)

        if open_editor:
            tmpfname = fs.get_temp_filepath()
            fs.copy_file(artifact_path, tmpfname)

            shell_cmd = shlex.split(config["EDITOR"]) + [tmpfname]
            call(shell_cmd)
            fs.remove_file(tmpfname)
            sys.exit(0)

        # View File
        if fs.is_text_file(artifact_path):
            markers = get_template(artifact, config)
            viewer.view(artifact_path, markers, color=color_mode)
        else:
            opener.open_non_text_file(artifact_path)
    elif len(artifacts) > 1:
        print(
            "There is more than one artifact with that title, please specify a category")
    else:
        print(
            "There is no artifact with that name, please specify a correct artifact name")


def get_template(artifact: Artifact, config: Dict[str, str]) -> str:
    """"
    Get template for a specific artifact.

    Arguments:
    artifact        - the artifact to get the template for
    config:         - a configuration dictionary containing at least
                      the following keys:
                      PATH_KB_DEFAULT_TEMPLATE   - the file associated to the markers
                      PATH_KB_TEMPLATES          - the path where templates are stored

    Returns:
    A dictionary containing markers, where the key is a regex
    and the value is a string representing a color.
    """
    template = artifact.template or "default"
    if template == "default":
        markers = get_markers(config["PATH_KB_DEFAULT_TEMPLATE"])
    else:
        markers = get_markers(
            str(Path(*[config["PATH_KB_TEMPLATES"]] + template.split('/'))))
    return markers



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
import io
from setuptools import setup, find_packages


setup(name='kb-manager',
      version='0.1.7',
      description='A minimalist knowledge base manager',
      keywords='kb',
      author='gnc',
      author_email='nebbionegiuseppe@gmail.com',
      url='https://github.com/gnebbia/kb',
      download_url='https://github.com/gnebbia/kb/archive/v0.1.7.tar.gz',
      license='GPLv3',
      long_description=io.open(
          './docs/README.md', 'r', encoding='utf-8').read(),
      long_description_content_type="text/markdown",
      platforms='any',
      zip_safe=False,
      # http://pypi.python.org/pypi?%3Aaction=list_classifiers
      classifiers=[ 'Programming Language :: Python',
                    'Programming Language :: Python :: 3',
                    'Programming Language :: Python :: 3.6',
                    'Programming Language :: Python :: 3.7',
                    'Programming Language :: Python :: 3.8',
                    'Operating System :: OS Independent',
                   ],
      packages=find_packages(exclude=('tests',)),
      include_package_data=True,
      install_requires=["colored","toml","attr","attrs","gitpython"],
      python_requires='>=3.6',
      entry_points={
           'console_scripts':[
               'kb = kb.main:main',
           ]
      },
      )



# -*- encoding: utf-8 -*-
# kb v0.1.7
# A knowledge base organizer
# Copyright © 2020, gnc.
# See /LICENSE for licensing information.

"""
kb export command module

:Copyright: © 2020, gnc.
:License: GPLv3 (see /LICENSE).
"""

import time
from typing import Dict
import kb.filesystem as fs
import git


def sync(args: Dict[str, str], config: Dict[str, str]):
    """
    Synchronize the knowledge base with a remote repository.

    Arguments:
    args:           - a dictionary containing the following fields:
                      file -> a string representing the wished output
                        filename
    config:         - a configuration dictionary containing at least
                      the following keys:
                      PATH_KB           - the main path of KB
    """

    operation = args["operation"]

    if operation == "init":
        if is_local_git_repo_initialized(config["PATH_KB_GIT"]):
            print("Warning: a git repository already exists...")
            cancel_repo = input(
                "Do you want to re-init your repository ? [Type YES in that case] ")
            if cancel_repo == "YES":
                fs.remove_directory(config["PATH_KB_GIT"])
                git_init(config["PATH_KB"])
            else:
                print("Maybe you wanted to type \"kb sync push\" or \"kb sync pull\"!")
        else:
            git_init(config["PATH_KB"])
    elif operation == "push":
        if not is_local_git_repo_initialized(config["PATH_KB_GIT"]):
            print("Warning: You did not initialize any repository...")
            print("try: kb sync init")
            return
        git_push(config["PATH_KB"])
    elif operation == "pull":
        if not is_local_git_repo_initialized(config["PATH_KB_GIT"]):
            print("Warning: no local git repository has been instantiated!")
            cancel_repo = input(
                "Do you want to remove possible local kb files and sync from a remote repository? [Type YES in that case] ")
            if cancel_repo == "YES":
                try:
                    fs.remove_directory(config["PATH_KB"])
                except BaseException:
                    pass
                git_clone(config["PATH_KB"])
        else:
            git_pull(config["PATH_KB_GIT"])
    elif operation == "info":
        if not is_local_git_repo_initialized(config["PATH_KB_GIT"]):
            print("No local git repository has been instantiated!")
            return
        show_info(config["PATH_KB"])
    else:
        print("Error: No valid operation has been specified")
        return


def is_local_git_repo_initialized(git_path):
    if fs.is_directory(git_path):
        return True


def git_push(repo_path):
    try:
        kb_repo = git.Repo(repo_path)
        kb_repo.git.add('--all')
        timestamp = time.strftime("%d/%m/%Y-%H:%M:%S")
        kb_repo.index.commit("kb synchronization {ts}".format(ts=timestamp))
        kb_repo.remote(
            name='origin').push(
            refspec='{}:{}'.format(
                "main",
                "main"))
        print("Repository correctly synchronized to remote!")
    except BaseException:
        print('Some error occurred while pushing the code')
        print('Check your internet connection or the existence of the remote repository')


def git_pull(repo_path):
    try:
        kb_repo = git.Repo(repo_path)
        origin = kb_repo.remotes.origin
        origin.pull(origin.refs[0].remote_head)
        print("Repository correctly synchronized from remote!")
    except BaseException:
        print('Some error occurred while pulling the code')
        print('Check your internet connection or the existence of the remote repository')


def git_clone(repo_path):
    remote_repo_url = input(
        "Insert the URL of the remote repo (e.g., https://github/user/mykb): ")
    if remote_repo_url:
        git.Repo.clone_from(remote_repo_url, repo_path)
        print("Knowledge base correctly pulled from remote!")
    else:
        print("Error: Check your internet connection or provide a valid repository")


def git_init(repo_path):
    print("Create a remote empty repository on github/gitlab or other git provider...")
    remote_repo_url = input(
        "Insert the URL of the created empty remote repo (e.g., https://github/user/mykb): ")
    if remote_repo_url:
        local_repo = git.Repo.init(repo_path)
        remote = local_repo.create_remote("origin", url=remote_repo_url)
        local_repo.git.add('--all')
        timestamp = time.strftime("%d/%m/%Y-%H:%M:%S")
        local_repo.index.commit("kb synchronization {ts}".format(ts=timestamp))
        print("Initialization with remote may take time...")
        print("Please provide remote credentials and wait...")
        remote.push(refspec='{}:{}'.format("main", "main"))
        print("Remote repository correctly initialized!")
    else:
        print("Error: Provide a valid remote URL")


def show_info(repo_path):
    repo = git.cmd.Git(repo_path)
    print("Remote repository:")
    print(repo.execute(["git", "remote", "-v"]))



# -*- encoding: utf-8 -*-
# kb v0.1.7
# A knowledge base organizer
# Copyright © 2020, gnc.
# See /LICENSE for licensing information.

"""
kb search command module

:Copyright: © 2020, gnc.
:License: GPLv3 (see /LICENSE).
"""

from typing import Dict
import kb.db as db
import kb.initializer as initializer
import kb.printer.search as printer
import kb.history as history


def search(args: Dict[str, str], config: Dict[str, str]):
    """
    Search artifacts within the knowledge base of kb.

    Arguments:
    args:           - a dictionary containing the following fields:
                      query -> filter for the title field of the artifact
                      category -> filter for the category field of the artifact
                      tags -> filter for the tags field of the artifact
                      author -> filter for the author field of the artifact
                      status -> filter for the status field of the artifact
    config:         - a configuration dictionary containing at least
                      the following keys:
                      PATH_KB_DB        - the database path of KB
                      PATH_KB_DATA      - the data directory of KB
                      PATH_KB_HIST      - the history menu path of KB
                      EDITOR            - the editor program to call
    """
    # Check initialization
    initializer.init(config)

    tags_list = None
    if args["tags"] and args["tags"] != "":
        tags_list = args["tags"].split(';')

    conn = db.create_connection(config["PATH_KB_DB"])
    rows = db.get_artifacts_by_filter(
        conn,
        title=args["query"],
        category=args["category"],
        tags=tags_list,
        status=args["status"],
        author=args["author"])

    # rows.sort(key=lambda x: x[1])
    artifacts = sorted(rows, key=lambda x: x.title)

    # Write to history file
    history.write(config["PATH_KB_HIST"], artifacts)

    # Is full_identifier mode enabled?
    if args["full_identifier"]:
        printer.print_search_result_full_mode(artifacts)
        return

    # Print resulting list
    color_mode = not args["no_color"]
    if args["verbose"]:
        printer.print_search_result_verbose(artifacts, color_mode)
    else:
        printer.print_search_result(artifacts, color_mode)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# kb test suite
# Copyright © 2020, gnc.
# See /LICENSE for licensing information.

"""
To reload modules
import imp
imp.reload(kb)
imp.reload(kb.db)
imp.reload(kb.filesystem)
"""

import attr
import os
import sqlite3
import pytest
from pathlib import Path
import kb
import kb.db as db
import kb.filesystem as fs
from kb.entities.artifact import Artifact


@pytest.fixture()
def db_connect():
    conn = sqlite3.connect(str(Path("tests","data","mydb.db")))
    return conn

def _list_tables(conn):
    """
    Return the list of tables currently in the database
    at the moment kb is based on a single table
    called "artifacts", but this may change in the future.

    Arguments:
    conn            -   the sqlite3 Connection object

    Returns:
    A list of strings representing the list of tables
    in the kb database
    """
    with conn:
        cur = conn.cursor()
        sql_query = """SELECT name FROM sqlite_master
                    WHERE type='table'
                    ORDER BY name;"""
        cur.execute(sql_query)
        rows = cur.fetchall()
    return rows


def test_create_connection():
    ok_db_path = Path("tests","data",".kb","kb.db")
    assert db.create_connection(str(ok_db_path)) is not None
    

def test_create_connection_2():
    failing_db_path = Path("tests","data","non_existing","kb.db")
    assert db.create_connection(str(failing_db_path)) is None


def test_create_connection_3():
    ok_db_path = ":memory:"
    assert db.create_connection(ok_db_path) is not None


def test_list_tables_blank():
    db_path = Path("tests","data","empty.db")
    conn = db.create_connection(str(db_path))
    with conn:
        assert _list_tables(conn) == []


def test_list_tables():
    sql_create_table_query = """CREATE TABLE IF NOT EXISTS testtable1 (
                                    id integer PRIMARY KEY,
                                    author text
                                 );
                                CREATE TABLE IF NOT EXISTS testtable2 (
                                    id integer PRIMARY KEY,
                                    author text
                                );
                             """

    db_path = Path("tests","data","two_tables.db")

    conn = db.create_connection(str(db_path))
    with conn:
        try:
            c = conn.cursor()
            c.executescript(sql_create_table_query)
        except Exception as e:
            print(e)

        assert len(_list_tables(conn)) == 2


def test_create_table(db_connect):
    sql_db_create_query = """CREATE TABLE IF NOT EXISTS artifacts (
                                id integer PRIMARY KEY,
                                title text NOT NULL,
                                category text NOT NULL,
                                path text NOT NULL,
                                tags text,
                                status text,
                                author text,
                                template text);
                          """

    db_path = Path("tests","data","mydb.db")

    conn = db.create_connection(str(db_path))
    with conn:
        if conn is not None:
            db.create_table(conn, sql_db_create_query)
        else:
            print("Error! cannot create the database connection.")

        assert len(_list_tables(conn)) == 1


def test_create_table_2(db_connect):
    sql_db_create_query = """CREATE TABLE artifacts (
                                id integer PRIMARY KEY,
                                title text NOT NULL,
                                category text NOT NULL,
                                path text NOT NULL,
                                tags text,
                                status text,
                                author text,
                                template text);
                          """
    db_path = Path("tests","data","mydb2.db")

    conn = db.create_connection(str(db_path))
    with conn:
        if conn is not None:
            db.create_table(conn, sql_db_create_query)
            db.create_table(conn, sql_db_create_query)
        else:
            print("Error! cannot create the database connection.")

        assert len(_list_tables(conn)) == 1


def test_create_kb_database_table():
    db_path = Path("tests","data","mydb3.db")
    schema_version = 1
    db.create_kb_database(str(db_path), schema_version)

    conn = db.create_connection(str(db_path))
    with conn:
        try:
            c = conn.cursor()
        except Exception as e:
            print(e)

        # Make sure, the database has the correct number of tables
        kb_tables = _list_tables(conn)
        assert len(kb_tables) == 2
        assert kb_tables == [("artifacts",),("tags",)]



def test_insert_artifact():
    db_path = Path("tests","data","test_insert.db")
    schema_version = 1
    db.create_kb_database(str(db_path), schema_version)
    conn = db.create_connection(str(db_path))
    with conn:
        db.insert_artifact(conn, Artifact(id=None, path="pentest/smb", title="pentest_smb", category="procedure", 
                tags='pt;smb', status="OK", author="gnc"))
        db.insert_artifact(conn, Artifact(id=None, path="protocol/ftp", title="ftp", category="cheatsheet", 
                status="Draft", author="elektroniz"))

        kb_tables = _list_tables(conn)
        assert len(kb_tables) == 2
        assert kb_tables == [("artifacts",),("tags",)]

        sql = "SELECT * FROM artifacts;"
        cur = conn.cursor()
        cur.execute(sql)
    
        rows = cur.fetchall()
        print(rows)
        assert set(rows) == {(1, 'pentest_smb', 'procedure',
                            'pentest/smb', 'pt;smb', 'OK', 'gnc', None),
                            (2, 'ftp', 'cheatsheet', 'protocol/ftp', None,
                            'Draft', 'elektroniz', None)}



def test_is_artifact_existing():
    db_path = Path("tests","data","test_exist.db")
    schema_version = 1
    db.create_kb_database(str(db_path), schema_version)
    conn = db.create_connection(str(db_path))
    with conn:
        db.insert_artifact(conn, Artifact(id=None, path="pentest/smb", title="pentest_smb",
                category="procedure", tags='pt;smb', status="OK", author="gnc"))
        db.insert_artifact(conn, Artifact(id=None, path="protocol/ftp", title="ftp",
                category="cheatsheet", status="Draft", author="elektroniz"))
        
        assert db.is_artifact_existing(conn,title="pentest_smb",
                                    category="procedure")
        assert db.is_artifact_existing(conn,title="ftp",
                                    category="cheatsheet")
        assert not db.is_artifact_existing(conn,title="pentest_smb",
                                        category="nonexist")
        assert not db.is_artifact_existing(conn,title="nonexist",
                                        category="procedure")
        assert not db.is_artifact_existing(conn,title="",
                                        category="cheatsheet")
        assert not db.is_artifact_existing(conn,title="", category="")


def test_delete_artifact_by_id():
    db_path = Path("tests","data","test_id.db")

    schema_version = 1
    db.create_kb_database(str(db_path), schema_version)
    conn = db.create_connection(str(db_path))
    with conn:
        db.insert_artifact(conn, Artifact(id=None, path="pentest/smb", title="pentest_smb",
                category="procedure", tags='pt;smb', status="OK", author="gnc"))
        db.insert_artifact(conn, Artifact(id=None, path="protocol/ftp", title="ftp",
                category="cheatsheet", status="Draft", author="elektroniz"))
        
        db.delete_artifact_by_id(conn, 1)

        sql = "SELECT * FROM artifacts;"
        cur = conn.cursor()
        cur.execute(sql)
    
        rows = cur.fetchall()
        assert len(rows) == 1
        assert set(rows) == {(2, 'ftp', 'cheatsheet', 'protocol/ftp', None,
                            'Draft', 'elektroniz', None)}

        db.delete_artifact_by_id(conn,2)

        sql = "SELECT * FROM artifacts;"
        cur = conn.cursor()
        cur.execute(sql)
    
        rows = cur.fetchall()
        assert len(rows) == 0
    # try:
    #     os.unlink(db_path)
    # except FileNotFoundError:
    #     pass


def test_delete_artifact_by_name():
    db_path = Path("tests","data","test_name.db")

    schema_version = 1
    db.create_kb_database(str(db_path), schema_version)
    conn = db.create_connection(str(db_path))
    with conn:
        db.insert_artifact(conn, Artifact(id=None, path="pentest/smb", title="pentest_smb",
                category="procedure", tags='pt;smb', status="OK", author="gnc"))
        db.insert_artifact(conn, Artifact(id=None, path="protocol/ftp", title="ftp",
                category="cheatsheet", status="Draft", author="elektroniz"))

        db.delete_artifact_by_name(conn, title="pentest_smb", category="")
        sql = "SELECT * FROM artifacts;"
        cur = conn.cursor()
        cur.execute(sql)
    
        rows = cur.fetchall()
        assert len(rows) == 2

        db.delete_artifact_by_name(conn, title="pentest_smb", category="procedure")
        sql = "SELECT * FROM artifacts;"
        cur = conn.cursor()
        cur.execute(sql)
    
        rows = cur.fetchall()
        assert len(rows) == 1
        assert set(rows) == {(2, 'ftp', 'cheatsheet', 'protocol/ftp', None,
                            'Draft', 'elektroniz',None)}
    # try:
    #     os.unlink(db_path)
    # except FileNotFoundError:
    #     pass


def test_get_artifacts_by_tags():
    db_path = Path("tests","data","kb_art_tags.db")
    conn = db.create_connection(str(db_path))
    with conn:
        schema_version = 1
        db.create_kb_database(str(db_path), schema_version)
        db.insert_artifact(conn, Artifact(id=None, path="cheatsheet/pentest_smb", title="pentest_smb",
                category="procedure", tags='pt;smb', status="ok", author="gnc"))
        db.insert_artifact(conn, Artifact(id=None, path="guides/ftp", title="ftp", category="cheatsheet", 
                status="draft", author="elektroniz"))
        db.insert_artifact(conn, Artifact(id=None, path="guides/http", title="http", category="cheatsheet", 
                status="OK", author="elektroniz"))
        db.insert_artifact(conn, Artifact(id=None, path="guides/irc", title="irc", category="cheatsheet", 
                tags="protocol", status="draft", author="elektroniz"))
        db.insert_artifact(conn, Artifact(id=None, path="cheatsheet/pentest_ftp", title="pentest_ftp", category="cheatsheet", 
                tags="pt", status="draft", author="elektroniz"))

        rows = db.get_artifacts_by_tags(conn, tags=["pt"], is_strict=False)
        assert len(rows) == 2

        rows = db.get_artifacts_by_tags(conn, tags=["p"], is_strict=False)
        assert len(rows) == 3

        rows = db.get_artifacts_by_tags(conn, tags=["pt"], is_strict=True)
        assert len(rows) == 2


def test_get_artifacts_by_title():
    db_path = Path("tests","data","kb_filter_title.db")
    conn = db.create_connection(str(db_path))
    with conn:
        schema_version = 1
        db.create_kb_database(str(db_path), schema_version)
        db.insert_artifact(conn, Artifact(id=None, path="cheatsheet/pentest_smb", title="pentest_smb",
                category="procedure", tags='pt;smb', status="ok", author="gnc"))
        db.insert_artifact(conn, Artifact(id=None, path="guides/ftp", title="ftp", category="cheatsheet", 
                status="draft", author="elektroniz"))
        db.insert_artifact(conn, Artifact(id=None, path="guides/http", title="http", category="cheatsheet", 
                status="OK", author="elektroniz"))
        db.insert_artifact(conn, Artifact(id=None, path="guides/irc", title="irc", category="cheatsheet", 
                tags="protocol", status="draft", author="elektroniz"))
        db.insert_artifact(conn, Artifact(id=None, path="cheatsheet/pentest_ftp", title="pentest_ftp", category="cheatsheet", 
                tags="pt", status="draft", author="elektroniz"))

        rows = db.get_artifacts_by_title(conn, query_string="", is_strict=False)
        assert len(rows) == 5
        
        rows = db.get_artifacts_by_title(conn, query_string="", is_strict=True)
        assert len(rows) == 0


def test_get_artifacts_by_category():
    db_path = Path("tests","data","kb_filter_cat.db")

    conn = db.create_connection(str(db_path))
    with conn:
        schema_version = 1
        db.create_kb_database(str(db_path), schema_version)

        db.insert_artifact(conn, Artifact(id=None, path="cheatsheet/pentest_smb", title="pentest_smb",
                category="procedure", tags='pt;smb', status="ok", author="gnc"))

        db.insert_artifact(conn, Artifact(id=None, path="guides/ftp", title="ftp",
                category="cheatsheet",
                status="draft", author="elektroniz"))

        db.insert_artifact(conn, Artifact(id=None, path="guides/http", title="http",
                category="cheatsheet", status="OK", author="elektroniz"))

        db.insert_artifact(conn, Artifact(id=None, path="guides/irc", title="irc",
                        category="cheatsheet", tags="protocol", status="draft",
                        author="elektroniz"))

        db.insert_artifact(conn, Artifact(id=None, path="cheatsheet/pentest_ftp", title="pentest_ftp",
                        category="cheatsheet", tags="pt", status="draft",
                        author="elektroniz"))

        db.insert_artifact(conn, Artifact(id=None, path="sheet/math", title="math_formulas",
                        category="sheet", tags="math", status="draft",
                        author="gnc"))

        db.insert_artifact(conn, Artifact(id=None, path="sheet/math2", title="geometry_formulas",
                        category="sheet", tags="math", status="draft",
                        author="gnc"))

        rows = db.get_artifacts_by_category(conn, query_string="", is_strict=False)
        assert len(rows) == 7

        rows = db.get_artifacts_by_category(conn, query_string="", is_strict=True)
        assert len(rows) == 0

        rows = db.get_artifacts_by_category(conn, query_string="sheet", is_strict=True)
        assert len(rows) == 2


def test_get_artifacts_by_filter():
    db_path = Path("tests","data","kb_filter.db")
    conn = db.create_connection(str(db_path))
    with conn:
        schema_version = 1
        db.create_kb_database(str(db_path), schema_version)

        db.insert_artifact(conn, Artifact(id=None, path="", title="pentest_smb",
                category="procedure", tags='pt;smb', status="ok", 
                author="gnc"))

        db.insert_artifact(conn, Artifact(id=None, path="", title="ftp",
                category="cheatsheet", tags="protocol", status="draft",
                author="elektroniz"))

        db.insert_artifact(conn, Artifact(id=None, path="", title="pentest_ftp",
                category="procedure", tags="pt;ftp", status="draft",
                author="elektroniz"))

        db.insert_artifact(conn, Artifact(id=None, path="general/CORS", title="CORS",
                category="general", tags="web", status="draft",
                author="elektroniz"))

        rows = db.get_artifacts_by_filter(conn, title="pentest",
                                        category="cheatsheet",
                                        tags=["pt"], is_strict=False)

        assert len(rows) == 0

        rows = db.get_artifacts_by_filter(conn, category="procedure",
                                        tags=["pt"], is_strict=False)
        
        assert sorted(list(set(rows)), key=lambda i: i.id) == [Artifact(1,"pentest_smb","procedure","procedure/pentest_smb","pt;smb","ok","gnc", None),
                            Artifact(3,"pentest_ftp","procedure","procedure/pentest_ftp","pt;ftp","draft", "elektroniz", None)]


        rows = db.get_artifacts_by_filter(conn, title="OR")
        assert set(rows) == {Artifact(4,"CORS","general", "general/CORS", "web", "draft","elektroniz", None)}


        rows = db.get_artifacts_by_filter(conn, category="cheatsheet",
                                        is_strict=False)
        assert set(rows) == {Artifact(2,"ftp","cheatsheet","cheatsheet/ftp","protocol", "draft", "elektroniz", None)}


        rows = db.get_artifacts_by_filter(conn, category="sheet",
                                        is_strict=False)
        assert set(rows) == {Artifact(2,"ftp","cheatsheet","cheatsheet/ftp","protocol", "draft", "elektroniz", None)}

        rows = db.get_artifacts_by_filter(conn, category="cheatsheet",
                                        is_strict=True)
        assert set(rows) == {Artifact(2,"ftp","cheatsheet","cheatsheet/ftp","protocol", "draft", "elektroniz", None)}

        rows = db.get_artifacts_by_filter(conn, category="sheet",
                                        is_strict=True)
        assert len(rows) == 0



# -*- encoding: utf-8 -*-
# kb v0.1.7
# A knowledge base organizer
# Copyright 2020, Giuseppe Nebbione.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
A knowledge base organizer

:Copyright: © 2020, gnc.
:License: GPLv3 (see /LICENSE).
"""

__title__ = 'kb'
__version__ = '0.1.7'
__author__ = 'gnc'
__license__ = 'GPLv3'
__docformat__ = 'restructuredtext en'

__all__ = ()

# import gettext
# G = gettext.translation('kb', '/usr/share/locale', fallback='C')
# _ = G.gettext
